#include "string.h"
#include "stdio.h"
#include "stdlib.h"

#include "SDL.h"
#include "SDL_ttf.h"
#include "SDL_mixer.h"

#include "BList.h"

#include "MOG2_auxiliar.h"

#include "samples.h"
#include "animscript.h"
#include "mog2.h"

#ifdef KITSCHY_DEBUG_MEMORY 
#include "debug_memorymanager.h"
#endif

AnimSCRIPT::AnimSCRIPT(void)
{
	command=ANIM_CMD_NONE;
	ipar=0;
	fpar.frame=0;
	fpar.p_offset_p=0;
	fpar.p_offset_o=0;
	fpar.sample=0;
	fpar.sample_volume=100;
	sample=0;
	sample_volume=100;
	lastvalidtile=ANIM_INVALID;
} /* AnimSCRIPT::AnimSCRIPT */ 


AnimSCRIPT::AnimSCRIPT(AnimSCRIPT &s)
{
	command=s.command;
	ipar=s.ipar;
	fpar=s.fpar;
	sample=s.sample;
	lastvalidtile=s.lastvalidtile;
	fparlist.Copy(s.fparlist);
	aparlist.Copy(s.aparlist);
} /* AnimSCRIPT::AnimSCRIPT */ 


AnimSCRIPT::~AnimSCRIPT(void)
{
	sample=0;
} /* AnimSCRIPT::~AnimSCRIPT */ 


char *AnimSCRIPT::generate_string(void)
{
	char *str=0;

	switch(command) {
	case ANIM_CMD_NONE:
		str=new char[strlen("NONE")+1];
		sprintf(str,"NONE");
		break;
	case ANIM_CMD_SINGLE:
		{
			char *str2,*str3;
			str=new char[strlen("SINGLE ( 000 , 000 )")+1+16];
			if (fpar.sample!=0) encode_string(fpar.sample->name,&str3);
						   else encode_string(0,&str3);
			sprintf(str,"SINGLE ( %i , %i %i %i %s %i )",ipar,fpar.frame,fpar.p_offset_p,fpar.p_offset_o,str3,fpar.sample_volume);
			str2=new char[strlen(str)+1];
			strcpy(str2,str);
			delete []str3;
			delete []str;
			str=str2;
		}
		break;
	case ANIM_CMD_SEQ:
		{
			int pos;
			List<AnimFRAME> l;
			AnimFRAME *f;
			char *str2,*str3;

			str=new char[strlen("SEQ ( 000 , [ ] )")+6*fparlist.Length()*24+1];
			sprintf(str,"SEQ ( %i , [ ",ipar);
			pos=strlen(str);

			l.Instance(fparlist);
			l.Rewind();
			while(l.Iterate(f)) {
				if (f->sample!=0) encode_string(f->sample->name,&str3);
					 	     else encode_string(0,&str3);
				sprintf(str+pos,"%i %i %i %s %i , ",f->frame,f->p_offset_p,f->p_offset_o,str3,f->sample_volume);
				delete []str3;
				pos=strlen(str);
			} /* while */ 
			if (!l.EmptyP()) pos-=2;
			sprintf(str+pos,"] )");
			str2=new char[strlen(str)+1];
			strcpy(str2,str);
			delete  []str;
			str=str2;
		}
		break;
	case ANIM_CMD_LOOP:
		{
			int pos;
			List<AnimFRAME> l;
			AnimFRAME *f;
			char *str2,*str3;

			str=new char[strlen("LOOP ( 000 , [ ] )")+6*fparlist.Length()*24+1];
			sprintf(str,"LOOP ( %i , [ ",ipar);
			pos=strlen(str);

			l.Instance(fparlist);
			l.Rewind();
			while(l.Iterate(f)) {
				if (f->sample!=0) encode_string(f->sample->name,&str3);
					 	     else encode_string(0,&str3);
				sprintf(str+pos,"%i %i %i %s %i , ",f->frame,f->p_offset_p,f->p_offset_o,str3,f->sample_volume);
				delete []str3;
				pos=strlen(str);
			} /* while */ 
			if (!l.EmptyP()) pos-=2;
			sprintf(str+pos,"] )");
			str2=new char[strlen(str)+1];
			strcpy(str2,str);
			delete  []str;
			str=str2;
		}		
		break;
	case ANIM_CMD_GSEQ:
		{
			List<AnimSCRIPT> l;
			AnimSCRIPT *as;
			List<char> ls;
			char *s;
			int len=0,pos;

			l.Instance(aparlist);
			l.Rewind();
			while(l.Iterate(as)) {
				s=as->generate_string();
				ls.Add(s);
				len+=strlen(s);
			} /* while */ 
			if (!l.EmptyP()) {
				len+=l.Length()*3;
			} /* if */ 
			len+=strlen("GSEQ (  )")+1;

			str=new char[len];

			sprintf(str,"GSEQ ( ");
			pos=strlen("GSEQ ( ");

			ls.Rewind();
			while(!ls.EmptyP()) {
				s=ls.ExtractIni();
				sprintf(str+pos,s);
				pos+=strlen(s);
				if (!ls.EmptyP()) {
					sprintf(str+pos," , ");
					pos+=3;
				} /* if */ 
				delete []s;
			} /* while */ 
			sprintf(str+pos," )");
		}
		break;
	case ANIM_CMD_GLOOP:
		{
			List<AnimSCRIPT> l;
			AnimSCRIPT *as;
			List<char> ls;
			char *s;
			int len=0,pos;

			l.Instance(aparlist);
			l.Rewind();
			while(l.Iterate(as)) {
				s=as->generate_string();
				ls.Add(s);
				len+=strlen(s);
			} /* while */ 
			if (!l.EmptyP()) {
				len+=l.Length()*3;
			} /* if */ 
			len+=strlen("GLOOP (  )")+1;

			str=new char[len];

			sprintf(str,"GLOOP ( ");
			pos=strlen("GLOOP ( ");

			ls.Rewind();
			while(!ls.EmptyP()) {
				s=ls.ExtractIni();
				sprintf(str+pos,s);
				pos+=strlen(s);
				if (!ls.EmptyP()) {
					sprintf(str+pos," , ");
					pos+=3;
				} /* if */ 
				delete []s;
			} /* while */ 
			sprintf(str+pos," )");
		}
		break;
	} /* switch */ 

	return str;
} /* AnimSCRIPT::generate_string */ 


bool AnimSCRIPT::load(FILE *fp,MOG2 *game)
{
	char tmp2[256],*sample_name;
	char tmp[80];

	if (1!=fscanf(fp,"%s",tmp)) return false;

	if (strcmp(tmp,"NONE")==0) {
		command=ANIM_CMD_NONE;
	} else if (strcmp(tmp,"SINGLE")==0) {
		command=ANIM_CMD_SINGLE;
		if (9!=fscanf(fp,"%s %i %s %i %i %i %s %i %s",tmp,&ipar,tmp,&fpar.frame,&fpar.p_offset_p,&fpar.p_offset_o,tmp2,&fpar.sample_volume,tmp)) return false;
		decode_string(tmp2,&sample_name);
		fpar.sample=0;
		fpar.sample_volume=100;
		{
			SAMPLE *s2;
			List<SAMPLE> l;

			l.Instance(game->samples);
			l.Rewind();
			while(l.Iterate(s2)) {
				if (s2->cmp(sample_name)) fpar.sample=s2;
			} /* while */ 
		}
		delete []sample_name;
	} else if (strcmp(tmp,"SEQ")==0) {
		bool end=false;
		command=ANIM_CMD_SEQ;
		if (4!=fscanf(fp,"%s %i %s %s",tmp,&ipar,tmp,tmp)) return false;
		while(!end) {
			AnimFRAME *f;
			f=new AnimFRAME();
			f->sample=0;
			f->sample_volume=100;
			if (5==fscanf(fp,"%i %i %i %s %i ",&f->frame,&f->p_offset_o,&f->p_offset_p,tmp2,&f->sample_volume)) {
				decode_string(tmp2,&sample_name);
				f->sample=0;
				{
					SAMPLE *s2;
					List<SAMPLE> l;

					l.Instance(game->samples);
					l.Rewind();
					while(l.Iterate(s2)) {
						if (s2->cmp(sample_name)) f->sample=s2;
					} /* while */ 
				}
				delete []sample_name;
				fparlist.Add(f);
			} else {
				delete f;
			} /* if */ 
			if (1!=fscanf(fp,"%s",tmp)) return false;
			if (strcmp(tmp,"]")==0) end=true;
		} /* while */ 
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else if (strcmp(tmp,"LOOP")==0) {
		bool end=false;
		command=ANIM_CMD_LOOP;
		if (4!=fscanf(fp,"%s %i %s %s",tmp,&ipar,tmp,tmp)) return false;
		while(!end) {
			AnimFRAME *f;
			f=new AnimFRAME();
			f->sample=0;
			f->sample_volume=100;
			if (5==fscanf(fp,"%i %i %i %s %i",&f->frame,&f->p_offset_o,&f->p_offset_p,tmp2,&f->sample_volume)) {
				decode_string(tmp2,&sample_name);
				f->sample=0;
				{
					SAMPLE *s2;
					List<SAMPLE> l;

					l.Instance(game->samples);
					l.Rewind();
					while(l.Iterate(s2)) {
						if (s2->cmp(sample_name)) f->sample=s2;
					} /* while */ 
				}
				delete []sample_name;
				fparlist.Add(f);
			} else {
				delete f;
			} /* if */ 
			if (1!=fscanf(fp,"%s",tmp)) return false;
			if (strcmp(tmp,"]")==0) end=true;
		} /* while */ 
		if (1!=fscanf(fp,"%s",tmp)) return false;	
	} else if (strcmp(tmp,"GSEQ")==0) {
		bool end=false;
		command=ANIM_CMD_GSEQ;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		while(!end) {
			AnimSCRIPT *as;
			as=new AnimSCRIPT();
			if (as->load(fp,game)) {
				aparlist.Add(as);
				if (1!=fscanf(fp,"%s",tmp)) return false;
				if (strcmp(tmp,")")==0) end=true;
			} else {
				end=true;
			} /* if */ 
		} /* while */ 
	} else if (strcmp(tmp,"GLOOP")==0) {
		bool end=false;
		command=ANIM_CMD_GLOOP;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		while(!end) {
			AnimSCRIPT *as;
			as=new AnimSCRIPT();
			if (as->load(fp,game)) {
				aparlist.Add(as);
				if (1!=fscanf(fp,"%s",tmp)) return false;
				if (strcmp(tmp,")")==0) end=true;
			} else {
				end=true;
			} /* if */ 
		} /* while */ 
	} else {
		return false;
	} /* if */ 
	return true;
} /* AnimSCRIPT::load */ 


bool AnimSCRIPT::save(FILE *fp)
{
	char *str;

	str=generate_string();

	fprintf(fp,"%s\n",str);

	delete []str;
	return true;
} /* AnimSCRIPT::save */ 




void AnimSCRIPT::init(void)
{
	tile=ANIM_INVALID;
	lastvalidtile=tile;
	sample=0;
	sample_volume=100;
	timmer=0;
	fparlist.Rewind();
	aparlist.Rewind();

	switch(command) {
	case ANIM_CMD_NONE:
		break;
	case ANIM_CMD_SINGLE:
		tile=fpar.frame;
		sample=fpar.sample;
		break;
	case ANIM_CMD_SEQ:
	case ANIM_CMD_LOOP:
		if (!fparlist.EmptyP()) {
			tile=(fparlist.GetObj())->frame;
			sample=(fparlist.GetObj())->sample;
		} /* if */ 

		break;
	case ANIM_CMD_GSEQ:
	case ANIM_CMD_GLOOP:
		if (!aparlist.EmptyP()) {
			aparlist.GetObj()->init();
			tile=aparlist.GetObj()->get_tile();
			sample=aparlist.GetObj()->get_sample();
		} /* if */ 
		break;
	} /* switch */ 

	if (tile!=ANIM_INVALID) lastvalidtile=tile;
} /* AnimSCRIPT::init */ 


bool AnimSCRIPT::cycle(void)
{
	timmer++;
	sample=0;
	sample_volume=100;

	switch(command) {
	case ANIM_CMD_NONE:
		{
			tile=ANIM_INVALID;
			p_offset_p=p_offset_o=0;
			return true;
		}
		break;
	case ANIM_CMD_SINGLE:
		{
			if (timmer<ipar) {
				tile=fpar.frame;
				p_offset_p=fpar.p_offset_p;
				p_offset_o=fpar.p_offset_o;
				sample=fpar.sample;
				sample_volume=fpar.sample_volume;
			} else {
				tile=ANIM_INVALID;
				p_offset_p=p_offset_o=0;
				return true;
			} /* if */ 
		}
		break;
	case ANIM_CMD_SEQ:
		{
			if (timmer>=ipar) {
				fparlist.Next();
				if (!fparlist.EndP()) {
					sample=(fparlist.GetObj())->sample;
					sample_volume=(fparlist.GetObj())->sample_volume;
				} /* if */ 
				timmer=0;
			} /* if */ 

			if (!fparlist.EndP()) {
				tile=(fparlist.GetObj())->frame;
				p_offset_p=(fparlist.GetObj())->p_offset_p;
				p_offset_o=(fparlist.GetObj())->p_offset_o;
			} else {
				tile=ANIM_INVALID;
				p_offset_p=p_offset_o=0;
				return true;
			} /* if */ 
		}
		break;
	case ANIM_CMD_LOOP:
		{
			if (timmer>=ipar) {
				fparlist.Next();
				if (!fparlist.EndP()) {
					sample=(fparlist.GetObj())->sample;
					sample_volume=(fparlist.GetObj())->sample_volume;
				} else { 
					if (!fparlist.EmptyP()) {
						sample=fparlist[0]->sample;
						sample_volume=fparlist[0]->sample_volume;
					} /* if */ 
				} /* if */ 
				timmer=0;
			} /* if */ 

			if (!fparlist.EndP()) {
				tile=(fparlist.GetObj())->frame;
				p_offset_p=(fparlist.GetObj())->p_offset_p;
				p_offset_o=(fparlist.GetObj())->p_offset_o;
			} else {
				fparlist.Rewind();
				tile=(fparlist.GetObj())->frame;
				p_offset_p=(fparlist.GetObj())->p_offset_p;
				p_offset_o=(fparlist.GetObj())->p_offset_o;
			} /* if */ 
		}		
		break;
	case ANIM_CMD_GSEQ:
		{
			if (!aparlist.EmptyP()) {
				if (!aparlist.EndP()) {
					if (aparlist.GetObj()->cycle()) {
						aparlist.Next();
						if (!aparlist.EndP()) {
							aparlist.GetObj()->init();
							sample=aparlist.GetObj()->get_sample();
							sample_volume=aparlist.GetObj()->get_sample_volume();
						} /* if */ 
					} /* if */ 
				} /* if */ 

				if (aparlist.EndP()) {
					tile=ANIM_INVALID;
					p_offset_p=p_offset_o=0;
					return true;
				} else {
					tile=aparlist.GetObj()->get_tile();
				} /* if */ 
			} else {
				tile=ANIM_INVALID;
				p_offset_p=p_offset_o=0;
				return true;
			} /* if */ 

		}
		break;
	case ANIM_CMD_GLOOP:
		{
			if (!aparlist.EmptyP()) {
				if (!aparlist.EndP()) {
					if (aparlist.GetObj()->cycle()) {
						aparlist.Next();
						if (!aparlist.EndP()) {
							aparlist.GetObj()->init();
							sample=aparlist.GetObj()->get_sample();
							sample_volume=aparlist.GetObj()->get_sample_volume();
						} /* if */ 
					} /* if */ 
				} /* if */ 

				if (aparlist.EndP()) {
					aparlist.Rewind();
					aparlist.GetObj()->init();
					tile=aparlist.GetObj()->get_tile();
					sample=aparlist.GetObj()->get_sample();
					sample_volume=aparlist.GetObj()->get_sample_volume();
				} else {
					tile=aparlist.GetObj()->get_tile();
				} /* if */ 
			} /* if */ 
		}
		break;
	} /* switch */ 

	if (tile!=ANIM_INVALID) lastvalidtile=tile;

	return false;
} /* AnimSCRIPT::cycle */ 


int AnimSCRIPT::get_tile(void)
{
	if (tile==ANIM_INVALID) return lastvalidtile;
	return tile;
} /* AnimSCRIPT::get_tile */ 


int AnimSCRIPT::get_p_offset_p(void)
{
	return p_offset_p;
} /* AnimSCRIPT::get_p_offset_p */ 


int AnimSCRIPT::get_p_offset_o(void)
{
	return p_offset_o;
} /* AnimSCRIPT::get_p_offset_o */ 

SAMPLE *AnimSCRIPT::get_sample(void)
{
	return sample;
} /* AnimSCRIPT::get_tile */ 

int AnimSCRIPT::get_sample_volume(void)
{
	return sample_volume;
} /* AnimSCRIPT::get_sample_volume */ 



void AnimSCRIPT::copy_state(AnimSCRIPT *s)
{
	tile=s->tile;
	timmer=s->timmer;

	command=s->command;
	ipar=s->ipar;
	fpar=s->fpar;

	if (fparlist.Length()!=s->fparlist.Length()) {
		fparlist.Delete();
		fparlist.Copy(s->fparlist);
	} else {
		List<AnimFRAME> l,l2;
		AnimFRAME *fp,*fp2;

		l.Instance(fparlist);
		l2.Instance(s->fparlist);
		l.Rewind();
		l2.Rewind();

		while(l.Iterate(fp) &&
			  l2.Iterate(fp2)) {
			*fp=*fp2;
		} /* while */ 
	} /* if */ 
	fparlist.Synchronize(&(s->fparlist));

	if (aparlist.Length()!=s->aparlist.Length()) {
		aparlist.Delete();
		aparlist.Copy(s->aparlist);
	} else {
		List<AnimSCRIPT> l,l2;
		AnimSCRIPT *ap,*ap2;

		l.Instance(aparlist);
		l2.Instance(s->aparlist);
		l.Rewind();
		l2.Rewind();

		while(l.Iterate(ap) &&
			  l2.Iterate(ap2)) {
			ap->copy_state(ap2);
		} /* while */ 
	} /* if */ 
	aparlist.Synchronize(&(s->aparlist));
	
} /* AnimSCRIPT::copy_state */ 

