#include "assert.h"

#include "stdio.h"
#include "string.h"
#include "stdlib.h"

#include "SDL.h"
#include "SDL_ttf.h"
#include "SDL_mixer.h"

#include "BList.h"

#include "auxiliar.h"

#include "samples.h"
#include "animscript.h"
#include "valuescript.h"
#include "conditionscript.h"
#include "actionscript.h"
#include "variables.h"
#include "mog2.h"
#include "object.h"
#include "activeobject.h"
#include "activeroom.h"
#include "activeMAP.h"
#include "activemog2.h"
#include "activeinterface.h"

#include "samples.h"

#ifdef KITSCHY_DEBUG_MEMORY 
#include "debug_memorymanager.h"
#endif

#define FALLING_THRESHOLD	8

void ActiveOBJECT::character_initialization(void)
{
	left_key=SDLK_LEFT;
	right_key=SDLK_RIGHT;
	up_key=SDLK_UP;
	down_key=SDLK_DOWN;
	fire1_key=SDLK_SPACE;
	fire2_key=SDLK_m;
	fire3_key=SDLK_n;
	fire4_key=SDLK_b;

	stopped_jumping=false;
	jumping_inertia=0;
	in_ladder=0;
	looking_right_in_ladder=true;

	continuous_state_change=false;

	invincible_script=0;
} /* ActiveOBJECT::character_initialization */ 


bool ActiveOBJECT::character_cycle(int *desired_tile,KEYBOARDSTATE **LVK,int n_LVKs)
{
	PLACED_OBJECT *old_in_ladder=in_ladder;
	KEYBOARDSTATE *keyboard=0;
	int	jump_key=object->jumping_key;
	bool anim_finished=false;

	if (gravity_ctnt<=0) gravity_ctnt=1;

	if (jump_key<0) {
		switch(jump_key) {
		case -1:jump_key=left_key;
			break;
		case -2:jump_key=right_key;
			break;
		case -3:jump_key=up_key;
			break;
		case -4:jump_key=down_key;
			break;
		case -5:jump_key=fire1_key;
			break;
		case -6:jump_key=fire2_key;
			break;
		case -7:jump_key=fire3_key;
			break;
		case -8:jump_key=fire4_key;
			break;
		} /* switch */ 
	} /* if */ 


	continuous_state_change=false;

	old_state=state;
	
	if (LVK!=0 && n_LVKs>LVK_ID) keyboard=LVK[LVK_ID];
	if (keyboard==0) return true;

	if (in_ladder!=0) check_collision=false;

	/* Animations: */ 
	if (state<CPOS_NSTATES) {
		int t,j;

		anim_finished=true;
		for(j=0;j<object->nparts;j++) {
			if (!state_anim[j][state]->cycle()) anim_finished=false;
			t=state_anim[j][state]->get_tile();
			desired_tile[j]=t;
			if (state_anim[j][state]->get_sample()!=0) {
				PLAYSAMPLE *ps;

				ps=new PLAYSAMPLE();
				ps->sample=state_anim[j][state]->get_sample();
				ps->volume=(state_anim[j][state]->get_sample_volume()*127)/100;
				ps->pan=0;
				ps->room=room;
				ps->placed=false;
				if (room!=0) {
					ps->placed=true;
					ps->x=x;
					ps->y=y;
				} /* if */ 
				get_game()->add_sample(ps);
			} /* if */ 
		} /* for */ 
	} /* if */ 

	/* Samples: */ 
	if (state<CPOS_NSTATES &&
		object->state_sample_command[state]!=0 &&
		object->state_sample[state]!=0) {
		PLAYSAMPLE *ps;
		int pan_x;

		switch(object->state_sample_command[state]) {
		case 1:
			if (state_timmer==0) {
				ps=new PLAYSAMPLE();
				ps->sample=object->state_sample[state];
				ps->volume=object->state_sample_volume[state];
				ps->pan=0;
				ps->room=room;
				ps->placed=false;
				if (room!=0) {
					ps->placed=true;
					ps->x=x;
					ps->y=y;
				} /* if */ 
				get_game()->add_sample(ps);
			} /* if */ 
			break;
		case 2:
			if (object->state_sample_command_parameter[state]==0 ||
				(state_timmer%object->state_sample_command_parameter[state])==0) {
				ps=new PLAYSAMPLE();
				ps->sample=object->state_sample[state];
				ps->volume=object->state_sample_volume[state];
				ps->pan=0;
				ps->room=room;
				ps->placed=false;
				if (room!=0) {
					ps->placed=true;
					ps->x=x;
					ps->y=y;
				} /* if */ 
				get_game()->add_sample(ps);
			} /* if */ 
			break;
		case 3:
			break;

		case 4:
			if (state_timmer==0) {
				pan_x=x;
				if (pan_x<0) pan_x=0;
				if (pan_x>room->room->get_dx()*TILESIZE) pan_x=room->room->get_dx()*TILESIZE;

				ps=new PLAYSAMPLE();
				ps->sample=object->state_sample[state];
				ps->volume=object->state_sample_volume[state];
				ps->pan=int( (long(pan_x)*200L) / long(room->room->get_dx()*TILESIZE) ) - 100;
				ps->room=room;
				ps->placed=false;
				if (room!=0) {
					ps->placed=true;
					ps->x=x;
					ps->y=y;
				} /* if */ 
				get_game()->add_sample(ps);			
			} /* if */ 
			break;
		case 5:
			if (object->state_sample_command_parameter[state]==0 ||
				(state_timmer%object->state_sample_command_parameter[state])==0) {
				pan_x=x;
				if (pan_x<0) pan_x=0;
				if (pan_x>room->room->get_dx()*TILESIZE) pan_x=room->room->get_dx()*TILESIZE;

				ps=new PLAYSAMPLE();
				ps->sample=object->state_sample[state];
				ps->volume=object->state_sample_volume[state];
				ps->pan=int( (long(pan_x)*200L) / long(room->room->get_dx()*TILESIZE) ) - 100;
				ps->room=room;
				ps->placed=false;
				if (room!=0) {
					ps->placed=true;
					ps->x=x;
					ps->y=y;
				} /* if */ 
				get_game()->add_sample(ps);			
			} /* if */ 
			break;
		case 6:
			break;

		} /* switch */ 
	} /* if */ 

	switch(state) {
	case CPOS_LANDINGRIGHT:
	case CPOS_STANDRIGHT:
		if (anim_finished && state==CPOS_LANDINGRIGHT) state=CPOS_STANDRIGHT;
		if (keyboard->keyboard[right_key]) state=CPOS_WALKINGRIGHT;
		if (keyboard->keyboard[left_key]) state=CPOS_WALKINGLEFT;
		if ((object->jump_states && keyboard->keyboard[jump_key] && !keyboard->old_keyboard[jump_key]) ||
			(object->stair_states && keyboard->keyboard[up_key] && !keyboard->old_keyboard[up_key])) {
			PLACED_OBJECT *ladder=search_ladder(0,0);

			if (ladder!=0 && (object->stair_states && keyboard->keyboard[up_key] && !keyboard->old_keyboard[up_key])) {
				in_ladder=ladder;
				looking_right_in_ladder=true;
				state=CPOS_UPSTAIRS;
				precision_move(0,-ladder_up_speed);
			} else {
				if (object->jump_states && keyboard->keyboard[jump_key] && !keyboard->old_keyboard[jump_key]) {
					jumping_inertia=0;
					state=CPOS_JUMPRIGHT;
					if (keyboard->keyboard[right_key]) jumping_inertia=walking_speed;
					if (keyboard->keyboard[left_key]) {
						jumping_inertia=-walking_speed;
						state=CPOS_JUMPLEFT;
					} /* if */ 
					stopped_jumping=false;
				} /* if */ 
			} /* if */ 
		} /* if */ 

		if (object->stair_states && keyboard->keyboard[down_key] && !keyboard->old_keyboard[down_key]) {
			PLACED_OBJECT *ladder=search_ladder(0,ladder_down_speed/100);

			if (ladder!=0) {
				in_ladder=ladder;
				looking_right_in_ladder=true;
				state=CPOS_DOWNSTAIRS;
				precision_move(0,ladder_down_speed);
			} /* if */ 
		} /* if */ 
		break;
	case CPOS_LANDINGLEFT:
	case CPOS_STANDLEFT:
		if (anim_finished && state==CPOS_LANDINGLEFT) state=CPOS_STANDLEFT;
		if (keyboard->keyboard[right_key]) state=CPOS_WALKINGRIGHT;
		if (keyboard->keyboard[left_key]) state=CPOS_WALKINGLEFT;
		if ((object->jump_states && keyboard->keyboard[jump_key] && !keyboard->old_keyboard[jump_key]) ||
			(object->stair_states && keyboard->keyboard[up_key] && !keyboard->old_keyboard[up_key])) {
			PLACED_OBJECT *ladder=search_ladder(0,0);

			if (ladder!=0 && (object->stair_states && keyboard->keyboard[up_key] && !keyboard->old_keyboard[up_key])) {
				in_ladder=ladder;
				looking_right_in_ladder=false;
				state=CPOS_UPSTAIRS;
				precision_move(0,-ladder_up_speed);
			} else {
				if (object->jump_states && keyboard->keyboard[jump_key] && !keyboard->old_keyboard[jump_key]) {
					jumping_inertia=0;
					state=CPOS_JUMPLEFT;
					if (keyboard->keyboard[left_key]) jumping_inertia=-walking_speed;
					if (keyboard->keyboard[right_key]) {
						jumping_inertia=walking_speed;
						state=CPOS_JUMPRIGHT;
					} /* if */ 
					stopped_jumping=false;
				} /* if */ 
			} /* if */ 
		} /* if */ 

		if (object->stair_states && keyboard->keyboard[down_key] && !keyboard->old_keyboard[down_key]) {
			PLACED_OBJECT *ladder=search_ladder(0,ladder_down_speed/100);

			if (ladder!=0) {
				in_ladder=ladder;
				looking_right_in_ladder=false;
				state=CPOS_DOWNSTAIRS;
				precision_move(0,ladder_down_speed);
			} /* if */ 
		} /* if */ 
		break;

	case CPOS_PAUSE:
		break;

	case CPOS_WALKINGRIGHT:
		precision_move(walking_speed,0);
		if (!keyboard->keyboard[right_key]) state=CPOS_STANDRIGHT;
		if (keyboard->keyboard[left_key] && !keyboard->old_keyboard[left_key]) state=CPOS_WALKINGLEFT;
		if ((object->jump_states && keyboard->keyboard[jump_key] && !keyboard->old_keyboard[jump_key]) ||
			(object->stair_states && keyboard->keyboard[up_key] && !keyboard->old_keyboard[up_key])) {
			PLACED_OBJECT *ladder=search_ladder(0,0);

			if (ladder!=0 && (object->stair_states && keyboard->keyboard[up_key] && !keyboard->old_keyboard[up_key])) {
				in_ladder=ladder;
				looking_right_in_ladder=true;
				state=CPOS_UPSTAIRS;
				precision_move(0,-ladder_up_speed);
			} else {
				if (object->jump_states && keyboard->keyboard[jump_key] && !keyboard->old_keyboard[jump_key]) {
					jumping_inertia=walking_speed;
					state=CPOS_JUMPRIGHT;
					stopped_jumping=false;
				} /* if */ 
			} /* if */ 
		} /* if */ 

		if (object->stair_states && keyboard->keyboard[down_key] && !keyboard->old_keyboard[down_key]) {
			PLACED_OBJECT *ladder=search_ladder(0,ladder_down_speed/100);

			if (ladder!=0) {
				in_ladder=ladder;
				looking_right_in_ladder=true;
				state=CPOS_DOWNSTAIRS;
				precision_move(0,ladder_down_speed);
			} /* if */ 
		} /* if */ 
		break;
	case CPOS_WALKINGLEFT:
		precision_move(-walking_speed,0);
		if (!keyboard->keyboard[left_key]) state=CPOS_STANDLEFT;
		if (keyboard->keyboard[right_key] && !keyboard->old_keyboard[right_key]) state=CPOS_WALKINGRIGHT;
		if ((object->jump_states && keyboard->keyboard[jump_key] && !keyboard->old_keyboard[jump_key]) ||
			(object->stair_states && keyboard->keyboard[up_key] && !keyboard->old_keyboard[up_key])) {
			PLACED_OBJECT *ladder=search_ladder(0,0);

			if (ladder!=0 && (object->stair_states && keyboard->keyboard[up_key] && !keyboard->old_keyboard[up_key])) {
				in_ladder=ladder;
				looking_right_in_ladder=false;
				state=CPOS_UPSTAIRS;
				precision_move(0,-ladder_up_speed);
			} else {
				if (object->jump_states && keyboard->keyboard[jump_key] && !keyboard->old_keyboard[jump_key]) {
					jumping_inertia=-walking_speed;
					state=CPOS_JUMPLEFT;
					stopped_jumping=false;
				} /* if */ 
			} /* if */ 
		} /* if */ 

		if (object->stair_states && keyboard->keyboard[down_key] && !keyboard->old_keyboard[down_key]) {
			PLACED_OBJECT *ladder=search_ladder(0,ladder_down_speed/100);

			if (ladder!=0) {
				in_ladder=ladder;
				looking_right_in_ladder=false;
				state=CPOS_DOWNSTAIRS;
				precision_move(0,ladder_down_speed);
			} /* if */ 
		} /* if */ 

		break;

	case CPOS_FALLINGRIGHT:
	case CPOS_FALLINGLEFT:
		if ((state_timmer/gravity_ctnt)==0) precision_move(0,200/gravity_ctnt);
		if ((state_timmer/gravity_ctnt)==1) precision_move(0,400/gravity_ctnt);
		if ((state_timmer/gravity_ctnt)==2) precision_move(0,600/gravity_ctnt);
		if ((state_timmer/gravity_ctnt)==3) precision_move(0,800/gravity_ctnt);
		if ((state_timmer/gravity_ctnt)==4) precision_move(0,800/gravity_ctnt);
		if ((state_timmer/gravity_ctnt)==5) precision_move(0,1000/gravity_ctnt);
		if ((state_timmer/gravity_ctnt)==6) precision_move(0,1000/gravity_ctnt);
		if ((state_timmer/gravity_ctnt)==7) precision_move(0,1000/gravity_ctnt);
		if ((state_timmer/gravity_ctnt)>=8) precision_move(0,1200/gravity_ctnt);
		break;

	case CPOS_UPSTAIRS:
		if (!keyboard->keyboard[up_key] && !keyboard->keyboard[down_key]) state=CPOS_QUIETUPSTAIRS;
		if (keyboard->keyboard[down_key] && !keyboard->old_keyboard[down_key]) state=CPOS_DOWNSTAIRS;
		if (keyboard->keyboard[down_key] && !keyboard->keyboard[up_key]) state=CPOS_DOWNSTAIRS;
		if (keyboard->keyboard[right_key] && !keyboard->keyboard[up_key]) state=CPOS_QUIETRIGHTSTAIRS;
		if (keyboard->keyboard[left_key] && !keyboard->keyboard[up_key]) state=CPOS_QUIETLEFTSTAIRS;
		if (keyboard->keyboard[up_key]) {
			PLACED_OBJECT *ladder=search_ladder(0,0);


			if (ladder!=0) {
				int ax=0,ay=0,oldax=0,olday=0;
				in_ladder=ladder;

				absolute_hotspot_coordinates(&oldax,&olday);
				precision_move(0,-ladder_up_speed);
				absolute_hotspot_coordinates(&ax,&ay);

				if (object->stop_at_each_floor &&
					!room->map->object_collision(this,ax,ay,true) &&
					room->map->object_collision(this,oldax,olday,true)) {
					if (looking_right_in_ladder) state=CPOS_STANDRIGHT;
											else state=CPOS_STANDLEFT;
					old_in_ladder=0;
					in_ladder=0;
				} /* if */ 
			} else {
				in_ladder=0;
				if (looking_right_in_ladder) state=CPOS_STANDRIGHT;
									    else state=CPOS_STANDLEFT;
			} /* if */ 

		} /* if */ 
		break;
	case CPOS_DOWNSTAIRS:
		if (!keyboard->keyboard[up_key] && !keyboard->keyboard[down_key]) state=CPOS_QUIETUPSTAIRS;
		if (keyboard->keyboard[up_key] && !keyboard->old_keyboard[up_key]) state=CPOS_UPSTAIRS;
		if (keyboard->keyboard[up_key] && !keyboard->keyboard[down_key]) state=CPOS_UPSTAIRS;
		if (keyboard->keyboard[right_key] && !keyboard->keyboard[left_key]) state=CPOS_QUIETRIGHTSTAIRS;
		if (keyboard->keyboard[left_key] && !keyboard->keyboard[right_key]) state=CPOS_QUIETLEFTSTAIRS;
		if (keyboard->keyboard[down_key]) {
			PLACED_OBJECT *ladder=search_ladder(0,ladder_down_speed/100);

			if (ladder!=0) {
				int ax=0,ay=0,oldax=0,olday=0;
				in_ladder=ladder;

				absolute_hotspot_coordinates(&oldax,&olday);
				precision_move(0,ladder_down_speed);
				absolute_hotspot_coordinates(&ax,&ay);

				if (object->stop_at_each_floor &&
					room->map->object_collision(this,ax,ay,true) &&
					!room->map->object_collision(this,oldax,olday,true)) {

					precision_move(0,-ladder_down_speed);
					if (looking_right_in_ladder) state=CPOS_STANDRIGHT;
											else state=CPOS_STANDLEFT;
					old_in_ladder=0;
					in_ladder=0;
				} /* if */ 
			} else {
				in_ladder=0;
				if (looking_right_in_ladder) state=CPOS_STANDRIGHT;
							 		    else state=CPOS_STANDLEFT;
			} /* if */ 
		} /* if */ 
		break;
	case CPOS_QUIETUPSTAIRS:
	case CPOS_QUIETDOWNSTAIRS:
	case CPOS_QUIETRIGHTSTAIRS:
	case CPOS_QUIETLEFTSTAIRS:
		{
			PLACED_OBJECT *ladder=search_ladder(0,0);

			if (keyboard->keyboard[right_key] && !keyboard->keyboard[left_key]) {
				int ax,ay;

				absolute_hotspot_coordinates(&ax,&ay);

				if (!room->map->object_collision(this,ax+4,ay-4,true) &&
					room->map->object_collision(this,ax+4,ay+5,true)) {
					state=CPOS_STANDRIGHT;
					old_in_ladder=0;
					in_ladder=0;
					precision_move(walking_speed,0);
				} else {
					state=CPOS_QUIETRIGHTSTAIRS;
				} /* if */ 
			} /* if */ 
			if (keyboard->keyboard[left_key] && !keyboard->keyboard[right_key]) {
				int ax,ay;

				absolute_hotspot_coordinates(&ax,&ay);

				if (!room->map->object_collision(this,ax-4,ay-4,true) &&
					room->map->object_collision(this,ax-4,ay+5,true)) {
					state=CPOS_STANDLEFT;
					old_in_ladder=0;
					in_ladder=0;
					precision_move(-walking_speed,0);
				} else {
					state=CPOS_QUIETLEFTSTAIRS;
				} /* if */ 
			} /* if */ 

			if (in_ladder!=0) {
				if (ladder!=0) {
					if (keyboard->keyboard[up_key] && !keyboard->keyboard[right_key] && !keyboard->keyboard[left_key]) {
						state=CPOS_UPSTAIRS;
					} /* if */ 
					if (keyboard->keyboard[down_key] && !keyboard->keyboard[right_key] && !keyboard->keyboard[left_key]) {
						state=CPOS_DOWNSTAIRS;
					} /* if */ 
				} else {
					in_ladder=0;
					if (looking_right_in_ladder) state=CPOS_STANDRIGHT;
											else state=CPOS_STANDLEFT;
				} /* if */ 
			} /* if */ 
		}
		break;

	case CPOS_JUMPRIGHT:
		precision_move(jumping_inertia,0);
		if (object->variable_jumping && (state_timmer/gravity_ctnt)<jumping_ctnt && !keyboard->keyboard[jump_key]) state_timmer=jumping_ctnt*gravity_ctnt;
		if (!stopped_jumping) {
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-11) precision_move(0,-1400/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-10) precision_move(0,-1200/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-9) precision_move(0,-1200/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-8) precision_move(0,-1200/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-7) precision_move(0,-1000/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-6) precision_move(0,-1000/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-5) precision_move(0,-1000/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-4) precision_move(0,-800/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-3) precision_move(0,-800/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-2) precision_move(0,-600/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-1) precision_move(0,-400/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt) precision_move(0,-200/gravity_ctnt);
		} /* if */ 

		if ((state_timmer/gravity_ctnt)>=jumping_ctnt+2) state=CPOS_JUMPFALLRIGHT;
		break;
	case CPOS_JUMPLEFT:
		precision_move(jumping_inertia,0);
		if (object->variable_jumping && (state_timmer/gravity_ctnt)<jumping_ctnt && !keyboard->keyboard[jump_key]) state_timmer=jumping_ctnt*gravity_ctnt;
		if (!stopped_jumping) {
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-11) precision_move(0,-1400/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-10) precision_move(0,-1200/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-9) precision_move(0,-1200/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-8) precision_move(0,-1200/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-7) precision_move(0,-1000/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-6) precision_move(0,-1000/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-5) precision_move(0,-1000/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-4) precision_move(0,-800/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-3) precision_move(0,-800/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-2) precision_move(0,-600/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt-1) precision_move(0,-400/gravity_ctnt);
			if ((state_timmer/gravity_ctnt)==jumping_ctnt) precision_move(0,-200/gravity_ctnt);
		} /* if */ 

		if ((state_timmer/gravity_ctnt)>=jumping_ctnt+2) state=CPOS_JUMPFALLLEFT;
		break;
	case CPOS_JUMPFALLRIGHT:
		precision_move(jumping_inertia,0);
		if (keyboard->keyboard[right_key] && object->controllable_jump) precision_move(walking_speed/(jumping_inertia==0 ? 1:2),0);
		if (keyboard->keyboard[left_key] && object->controllable_jump) {
			precision_move(-walking_speed/(jumping_inertia==0 ? 1:2),0);
			state=CPOS_JUMPFALLLEFT;
			continuous_state_change=true;
		} /* if */ 
		if (state_timmer>=3) {
			if (((state_timmer-3)/gravity_ctnt)==0) precision_move(0,200/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)==1) precision_move(0,400/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)==2) precision_move(0,600/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)==3) precision_move(0,800/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)==4) precision_move(0,800/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)==5) precision_move(0,1000/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)==6) precision_move(0,1000/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)==7) precision_move(0,1000/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)>=8) precision_move(0,1200/gravity_ctnt);
		} /* if */ 
		break;
	case CPOS_JUMPFALLLEFT:
		precision_move(jumping_inertia,0);
		if (keyboard->keyboard[right_key] && object->controllable_jump) {
			precision_move(walking_speed/(jumping_inertia==0 ? 1:2),0);
			state=CPOS_JUMPFALLRIGHT;
			continuous_state_change=true;
		} /* if */ 
		if (keyboard->keyboard[left_key] && object->controllable_jump) precision_move(-walking_speed/(jumping_inertia==0 ? 1:2),0);
		if (state_timmer>=3) {
			if (((state_timmer-3)/gravity_ctnt)==0) precision_move(0,200/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)==1) precision_move(0,400/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)==2) precision_move(0,600/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)==3) precision_move(0,800/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)==4) precision_move(0,800/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)==5) precision_move(0,1000/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)==6) precision_move(0,1000/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)==7) precision_move(0,1000/gravity_ctnt);
			if (((state_timmer-3)/gravity_ctnt)>=8) precision_move(0,1200/gravity_ctnt);
		} /* if */ 
		break;

	case CPOS_RECOVERRIGHT:
		if (state_timmer>=4) state=CPOS_STANDRIGHT;
		break;
	case CPOS_RECOVERLEFT:
		if (state_timmer>=4) state=CPOS_STANDLEFT;
		break;
	case CPOS_DEADRIGHT:
	case CPOS_DEADLEFT:
		{
			int j;
			bool end=true;

			for(j=0;j<object->nparts;j++) {
				if (ANIM_INVALID!=state_anim[j][state]->tile) end=false;
			} /* if */ 
			if (end) terminate=true;
		}
		break;
	} /* switch */ 

	if (invincible_script!=0) {
		if (invincible_script->cycle(this,room,room->map,get_game(),0,get_game()->app,&(state_object_foreign[object->n_states+object->n_s_states]),LOCATION_OBJECT,LVK,n_LVKs)) {
			delete invincible_script;
			invincible_script=0;
		} /* if */ 
	} /* if */ 

	/* Falling: */ 
	if (state==CPOS_LANDINGRIGHT ||
		state==CPOS_STANDRIGHT ||
		state==CPOS_WALKINGRIGHT ||
		state==CPOS_RECOVERRIGHT ||

		state==CPOS_LANDINGLEFT ||
		state==CPOS_STANDLEFT ||
		state==CPOS_WALKINGLEFT ||
		state==CPOS_RECOVERLEFT) {
		int ax,ay;
		int tmp_res=0,tmp_res2=0;

		absolute_hotspot_coordinates(&ax,&ay);

		tmp_res=room->map->object_collision(this,ax,ay,true)&CONS_SOLID_SURFACE;
		tmp_res2=room->map->object_collision(this,ax,ay+1,true);
		if ((tmp_res2&CONS_SOLID)==0 &&
			((tmp_res2&CONS_SOLID_SURFACE)==0 || tmp_res!=0)) {
			tmp_res2=room->map->object_collision(this,ax,ay+FALLING_THRESHOLD,true);
			if ((tmp_res2&CONS_SOLID)==0 &&
				((tmp_res2&CONS_SOLID_SURFACE)==0 || tmp_res!=0)) {
				if (state==CPOS_LANDINGRIGHT ||
					state==CPOS_STANDRIGHT ||
					state==CPOS_WALKINGRIGHT ||
					state==CPOS_RECOVERRIGHT) state=CPOS_FALLINGRIGHT;
										 else state=CPOS_FALLINGLEFT;
			} else {
				int i,max=1;
				for(i=1;i<FALLING_THRESHOLD;i++) {
					if (!room->map->object_collision(this,ax,ay+i,true)) max=i;
				} /* if */ 
				y+=max;
			} /* if */ 
		} /* if */ 
	} /* if */ 

	/* Death detection: */ 
	if (energy<=0 && state!=CPOS_DEADRIGHT && state!=CPOS_DEADLEFT) {
		int i=state;
		if (i==CPOS_STANDRIGHT ||
			i==CPOS_WALKINGRIGHT ||
			i==CPOS_FALLINGRIGHT ||
			i==CPOS_LANDINGRIGHT ||
			i==CPOS_RECOVERRIGHT ||
			i==CPOS_DEADRIGHT ||
			((i==CPOS_UPSTAIRS ||
			  i==CPOS_DOWNSTAIRS ||
			  i==CPOS_QUIETUPSTAIRS ||
			  i==CPOS_QUIETDOWNSTAIRS) &&
			 looking_right_in_ladder) ||
			i==CPOS_QUIETRIGHTSTAIRS ||
			i==CPOS_JUMPRIGHT ||
			i==CPOS_JUMPFALLRIGHT) {
			/* Looking right: */ 
			state=CPOS_DEADRIGHT;
//			special_state=-1;
		} else {
			/* Lookinf left: */ 
			state=CPOS_DEADLEFT;
//			special_state=-1;
		} /* if */ 
	} /* if */ 

	if (state==CPOS_UPSTAIRS ||
		state==CPOS_DOWNSTAIRS ||
		state==CPOS_QUIETUPSTAIRS ||
		state==CPOS_QUIETDOWNSTAIRS ||
		state==CPOS_QUIETRIGHTSTAIRS ||
		state==CPOS_QUIETLEFTSTAIRS) check_collision=false;

	if (state==old_state) state_timmer++;

	if (in_ladder!=0) {

		if (in_ladder->room==this->room->room) {
			int x1,y1,x2,y2;
			in_ladder->object->get_bbox(in_ladder->x,in_ladder->y,in_ladder->object->default_tile,&x1,&y1,&x2,&y2);
			x=(x1+x2)/2;

			precision_x=0;
		} /* if */ 
	} else {
		/* If the character just have exited a ladder on this frame, the new position have to	*/ 
		/* be also aligned with the ladder:														*/ 
		if (old_in_ladder!=0 && old_in_ladder->room==this->room->room) {
			int x1,y1,x2,y2;
			old_in_ladder->object->get_bbox(old_in_ladder->x,old_in_ladder->y,old_in_ladder->object->default_tile,&x1,&y1,&x2,&y2);
			x=(x1+x2)/2;

			precision_x=0;
		} /* if */ 
	} /* if */ 

	return true;
} /* ActiveOBJECT::character_cycle */ 


void ActiveOBJECT::character_postcycle(void)
{
	/* Samples: */ 
	if (old_state<CPOS_NSTATES &&
		state!=old_state) {
		PLAYSAMPLE *ps;
		int pan_x;

		switch(object->state_sample_command[old_state]) {
		case 1:
			break;
		case 2:
			break;
		case 3:
			ps=new PLAYSAMPLE();
			ps->sample=object->state_sample[old_state];
			ps->volume=object->state_sample_volume[old_state];
			ps->pan=0;
			ps->room=room;
			ps->placed=false;
			if (room!=0) {
				ps->placed=true;
				ps->x=x;
				ps->y=y;
			} /* if */ 
			get_game()->add_sample(ps);			
			break;

		case 4:
			break;
		case 5:
			break;
		case 6:
			pan_x=x;
			if (pan_x<0) pan_x=0;
			if (pan_x>room->room->get_dx()*TILESIZE) pan_x=room->room->get_dx()*TILESIZE;

			ps=new PLAYSAMPLE();
			ps->sample=object->state_sample[old_state];
			ps->volume=object->state_sample_volume[old_state];
			ps->pan=int( (long(pan_x)*200L) / long(room->room->get_dx()*TILESIZE) ) - 100;
			ps->room=room;
			ps->placed=false;
			if (room!=0) {
				ps->placed=true;
				ps->x=x;
				ps->y=y;
			} /* if */ 
			get_game()->add_sample(ps);			
			break;

		} /* switch */ 
	} /* if */ 

	if (state<CPOS_NSTATES && 
		state!=old_state) {
		int j;

		if (continuous_state_change) {
			state_timmer++;
			for(j=0;j<object->nparts;j++) {
				state_anim[j][state]->init();
				if (state_anim[j][state]->get_sample()!=0) {
					PLAYSAMPLE *ps;

					ps=new PLAYSAMPLE();
					ps->sample=state_anim[j][state]->get_sample();
					ps->volume=(state_anim[j][state]->get_sample_volume()*127)/100;
					ps->pan=0;
					ps->room=room;
					ps->placed=false;
					if (room!=0) {
						ps->placed=true;
						ps->x=x;
						ps->y=y;
					} /* if */ 
					get_game()->add_sample(ps);
				} /* if */ 
			} /* for */ 
		} else {
			state_timmer=0;
			for(j=0;j<object->nparts;j++) {
				state_anim[j][state]->init();
				if (state_anim[j][state]->get_sample()!=0) {
					PLAYSAMPLE *ps;

					ps=new PLAYSAMPLE();
					ps->sample=state_anim[j][state]->get_sample();
					ps->volume=(state_anim[j][state]->get_sample_volume()*127)/100;
					ps->pan=0;
					ps->room=room;
					ps->placed=false;
					if (room!=0) {
						ps->placed=true;
						ps->x=x;
						ps->y=y;
					} /* if */ 
					get_game()->add_sample(ps);
				} /* if */ 
			} /* for */ 
		} /* if */ 

	} /* if */ 
} /* ActiveOBJECT::character_postcycle */ 



void ActiveOBJECT::character_collision(void)
{
	int rx,ry;
	int candidate_x=old_x,candidate_y=old_y;
	int init_col=0;
	bool g_found=false;

	rx=room->room->get_x()*TILESIZE;
	ry=room->room->get_y()*TILESIZE;

	/* Check for collisions: */ 
	if (old_x!=x || old_y!=y ||
		old_scale!=absolute_scale() ||
		old_angle!=absolute_angle()) {

		if (room->map->object_collision(this,true)!=0) {
			bool found=false;
			int i,maxy=old_y,maxx=old_x;

			init_col=room->map->object_collision(this,old_x+rx,old_y+ry,true)&CONS_SOLID_SURFACE;
			if (x!=old_x) {
				found=false;
				if (x<old_x) {
					for(i=old_x;i>=x;i--) {
						if ((room->map->object_collision(this,i+rx,old_y+ry,true)&CONS_SOLID)==0) {
							found=true;
							maxx=i;
						} /* if */ 
					} /* if */ 
				} else {
					for(i=old_x;i<=x;i++) {
						if ((room->map->object_collision(this,i+rx,old_y+ry,true)&CONS_SOLID)==0) {
							found=true;
							maxx=i;
						} /* if */ 
					} /* if */ 
				} /* if */ 

				if (found) {
					g_found=true;
					if ((state==CPOS_JUMPRIGHT ||
						 state==CPOS_JUMPLEFT ||
						 state==CPOS_JUMPFALLRIGHT ||
						 state==CPOS_JUMPFALLLEFT) &&
						maxx!=x) jumping_inertia=0;
					candidate_x=maxx;
				} /* if */ 
			} /* if */ 

			if (y!=old_y) {
				found=false;
				i=old_y;
				if (y<old_y) {
					for(i=old_y;i>=y;i--) {
						if ((room->map->object_collision(this,candidate_x+rx,i+ry,true)&CONS_SOLID)==0) {
							maxy=i;
							found=true;
						} /* if */ 
					} /* for */ 
					if ((state==CPOS_JUMPRIGHT ||
						 state==CPOS_JUMPLEFT) &&
						found && maxy!=y) stopped_jumping=true;
				} else {
					for(i=old_y;i<=y;i++) {
						if ((room->map->object_collision(this,candidate_x+rx,i+ry,true)|init_col)==init_col) {
							maxy=i;
							found=true;
						} /* if */  
					} /* for */ 
					if (found && maxy!=y) {
						if (state==CPOS_FALLINGRIGHT ||
							state==CPOS_JUMPFALLRIGHT) state=CPOS_LANDINGRIGHT;
						if (state==CPOS_FALLINGLEFT ||
							state==CPOS_JUMPFALLLEFT) state=CPOS_LANDINGLEFT;
					} /* if */ 
				} /* if */ 
				if (found) {
					g_found=true;
					candidate_y=maxy;
				} else {
					g_found=false;
				} /* if */ 
			} /* if */ 
		} /* if */ 

		if ((room->map->object_collision(this,candidate_x+rx,candidate_y+ry,true)&CONS_SOLID)!=0 ||
			!g_found) {

			find_without_collision_place(8+abs(x-old_x),8+abs(y-old_y));
			if ((room->map->object_collision(this,x+rx,y+ry,true)&CONS_SOLID)!=0) {
				x=old_x;
				y=old_y;
			} /* if */ 
		} else {
			x=candidate_x;
			y=candidate_y;
		} /* if */ 
	} /* if */ 

} /* ActiveOBJECT::character_collision */ 


PLACED_OBJECT *ActiveOBJECT::search_ladder(int xoffs,int yoffs)
{
	int x1,y1,x2,y2;

	get_absbbox(&x1,&y1,&x2,&y2);

	return room->map->map->search_ladder(x1+xoffs,y1+yoffs,x2-x1,y2-y1);
} /* ActiveOBJECT::search_ladder */ 
