#include "string.h"
#include "stdio.h"
#include "stdlib.h"
#include "math.h"

#include "SDL.h"
#include "SDL_ttf.h"
#include "SDL_mixer.h"

#include "BList.h"
#include "auxiliar.h"
#include "MOG2_auxiliar.h"

#include "valuescript.h"
#include "conditionscript.h"
#include "variables.h"
#include "mog2.h"
#include "object.h"
#include "interfaces.h"
#include "activeobject.h"
#include "activeroom.h"
#include "activeMAP.h"
#include "activemog2.h"
#include "ActiveINTERFACE.h"
#include "ActiveMOG2Application.h"
#include "samples.h"

#ifdef KITSCHY_DEBUG_MEMORY 
#include "debug_memorymanager.h"
#endif



ValueSCRIPT::ValueSCRIPT(void)
{
	command=VALUE_CMD_INTEGER;
	ipar=0;
	spar=0;
	vlocation=VALUE_VAR_OBJECT;
	vname=0;
	vpar[0]=0;
	vpar[1]=0;
	vpar[2]=0;
	vpar[3]=0;
	vpar[4]=0;
	vpar[5]=0;

	cond=0;

	sample=0;

	dx=0;
	dy=0;
} /* ValueSCRIPT::ValueSCRIPT */ 


ValueSCRIPT::ValueSCRIPT(ValueSCRIPT &s)
{
	command=s.command;
	ipar=s.ipar;
	if (s.spar!=0) {
		spar=new char[strlen(s.spar)+1];
		strcpy(spar,s.spar);
	} else {
		spar=0;
	} /* if */ 

	vlocation=s.vlocation;

	if (s.vname!=0) {
		vname=new char[strlen(s.vname)+1];
		strcpy(vname,s.vname);
	} else {
		vname=0;
	} /* if */ 

	vpar[0]=0;
	vpar[1]=0;
	vpar[2]=0;
	vpar[3]=0;
	vpar[4]=0;
	vpar[5]=0;
	if (s.vpar[0]!=0) vpar[0]=new ValueSCRIPT(*s.vpar[0]);
	if (s.vpar[1]!=0) vpar[1]=new ValueSCRIPT(*s.vpar[1]);
	if (s.vpar[2]!=0) vpar[2]=new ValueSCRIPT(*s.vpar[2]);
	if (s.vpar[3]!=0) vpar[3]=new ValueSCRIPT(*s.vpar[3]);
	if (s.vpar[4]!=0) vpar[4]=new ValueSCRIPT(*s.vpar[4]);
	if (s.vpar[5]!=0) vpar[5]=new ValueSCRIPT(*s.vpar[5]);

	cond=0;
	if (s.cond!=0) cond=new ConditionSCRIPT(*s.cond);

	sample=s.sample;

	dx=0;
	dy=0;
} /* ValueSCRIPT::ValueSCRIPT */ 


ValueSCRIPT::~ValueSCRIPT(void)
{
	delete []spar;
	delete []vname;

	if (vpar[0]!=0) delete vpar[0];
	if (vpar[1]!=0) delete vpar[1];
	if (vpar[2]!=0) delete vpar[2];
	if (vpar[3]!=0) delete vpar[3];
	if (vpar[4]!=0) delete vpar[4];
	if (vpar[5]!=0) delete vpar[5];

	delete cond;
} /* ValueSCRIPT::~ValueSCRIPT */ 


char *ValueSCRIPT::generate_string(void)
{
	char *str=0;

	switch(command) {
	case VALUE_CMD_INTEGER:
		str=new char[strlen("INTEGER 000000")+1];
		sprintf(str,"INTEGER %i",ipar);
		break;
	case VALUE_CMD_KEY:
		str=new char[strlen("KEY 000000")+1];
		sprintf(str,"KEY %i",ipar);
		break;
	case VALUE_CMD_STRING:
		{
			char *s1;

			encode_string(spar,&s1);
			str=new char[strlen("STRING ")+strlen(s1)+1];
			sprintf(str,"STRING %s",s1);
			delete []s1;
		}
		break;
	case VALUE_CMD_VARIABLE:
		str=new char[strlen("VARIABLE 0 ")+strlen(vname)+1];
		sprintf(str,"VARIABLE %i %s",vlocation,vname);
		break;
	case VALUE_CMD_RANDOM:
		{
			char *s1,*s2;

			s1=vpar[0]->generate_string();
			s2=vpar[1]->generate_string();
			str=new char[strlen("RANDOM (  ,  )")+strlen(s1)+strlen(s2)+1];
			sprintf(str,"RANDOM ( %s , %s )",s1,s2);
			delete []s1;
			delete []s2;
		}
		break;
	case VALUE_CMD_ADD:
		{
			char *s1,*s2;

			s1=vpar[0]->generate_string();
			s2=vpar[1]->generate_string();
			str=new char[strlen("ADD (  ,  )")+strlen(s1)+strlen(s2)+1];
			sprintf(str,"ADD ( %s , %s )",s1,s2);
			delete []s1;
			delete []s2;
		}
		break;
	case VALUE_CMD_SUB:
		{
			char *s1,*s2;

			s1=vpar[0]->generate_string();
			s2=vpar[1]->generate_string();
			str=new char[strlen("SUB (  ,  )")+strlen(s1)+strlen(s2)+1];
			sprintf(str,"SUB ( %s , %s )",s1,s2);
			delete []s1;
			delete []s2;
		}
		break;
	case VALUE_CMD_MUL:
		{
			char *s1,*s2;

			s1=vpar[0]->generate_string();
			s2=vpar[1]->generate_string();
			str=new char[strlen("MUL (  ,  )")+strlen(s1)+strlen(s2)+1];
			sprintf(str,"MUL ( %s , %s )",s1,s2);
			delete []s1;
			delete []s2;
		}
		break;
	case VALUE_CMD_DIV:
		{
			char *s1,*s2;

			s1=vpar[0]->generate_string();
			s2=vpar[1]->generate_string();
			str=new char[strlen("DIV (  ,  )")+strlen(s1)+strlen(s2)+1];
			sprintf(str,"DIV ( %s , %s )",s1,s2);
			delete []s1;
			delete []s2;
		}
		break;
	case VALUE_CMD_ABS:
		{
			char *s1;

			s1=vpar[0]->generate_string();
			str=new char[strlen("ABS (  )")+strlen(s1)+1];
			sprintf(str,"ABS ( %s )",s1);
			delete []s1;
		}
		break;
	case VALUE_CMD_COUNT:
		{
			char *s1,*s2,*s3;

			s1=vpar[0]->generate_string();
			s2=vpar[1]->generate_string();
			s3=cond->generate_string();
			str=new char[strlen("COUNT (  ,  ,  )")+strlen(s1)+strlen(s2)+strlen(s3)+1];
			sprintf(str,"COUNT ( %s , %s , %s )",s1,s2,s3);
			delete []s1;
			delete []s2;
			delete []s3;
		}
		break;
	case VALUE_CMD_COUNTCHILDREN:
		{
			char *s1,*s2;

			s1=vpar[0]->generate_string();
			s2=cond->generate_string();
			str=new char[strlen("COUNTCHILDREN (  ,  )")+strlen(s1)+strlen(s2)+1];
			sprintf(str,"COUNTCHILDREN ( %s , %s )",s1,s2);
			delete []s1;
			delete []s2;
		}
		break;
	case VALUE_CMD_CONCATENATE:
		{
			char *s1,*s2;

			s1=vpar[0]->generate_string();
			s2=vpar[1]->generate_string();
			str=new char[strlen("CONCATENATE (  ,  )")+strlen(s1)+strlen(s2)+1];
			sprintf(str,"CONCATENATE ( %s , %s )",s1,s2);
			delete []s1;
			delete []s2;
		}
		break;
	case VALUE_CMD_REMOVE_SUBSTRING:
		{
			char *s1,*s2;

			s1=vpar[0]->generate_string();
			s2=vpar[1]->generate_string();
			str=new char[strlen("REMOVE_SUBSTRING (  ,  )")+strlen(s1)+strlen(s2)+1];
			sprintf(str,"REMOVE_SUBSTRING ( %s , %s )",s1,s2);
			delete []s1;
			delete []s2;
		}
		break;
	case VALUE_CMD_MAKE_VARIABLE:
		{
			char *s1,*s2;

			s1=vpar[0]->generate_string();
			s2=vpar[1]->generate_string();
			str=new char[strlen("MAKE_VARIABLE (  ,  )")+strlen(s1)+strlen(s2)+1];
			sprintf(str,"MAKE_VARIABLE ( %s , %s )",s1,s2);
			delete []s1;
			delete []s2;
		}
		break;
	case VALUE_CMD_KEYPRESS_ASCII:
		{
			str=new char[strlen("KEYPRESS_ASCII")+1];
			sprintf(str,"KEYPRESS_ASCII");
		}
		break;
	case VALUE_CMD_KEYPRESS_KEY:
		{
			str=new char[strlen("KEYPRESS_KEY")+1];
			sprintf(str,"KEYPRESS_KEY");
		}
		break;
	case VALUE_CMD_KEYPRESS_NAME:
		{
			char *s1;

			s1=vpar[0]->generate_string();
			str=new char[strlen("KEYPRESS_NAME (  )")+strlen(s1)+1];
			sprintf(str,"KEYPRESS_NAME ( %s )",s1);
			delete []s1;
		}
		break;
	case VALUE_CMD_SAMPLE:
		{
			char *s1;

			if (sample!=0) {
				encode_string(sample->name,&s1);
				str=new char[strlen("SAMPLE ")+strlen(s1)+1];
				sprintf(str,"SAMPLE %s",s1);
				delete []s1;
			} else {
				sprintf(str,"SAMPLE ?");
			} /* if */ 
		}
		break;
	case VALUE_CMD_GETNPLAYERS:
		{
			str=new char[strlen("GETNPLAYERS")+1];
			sprintf(str,"GETNPLAYERS");
		}
		break;
	case VALUE_CMD_GETPLAYERNAME:
		{
			char *s1;

			s1=vpar[0]->generate_string();
			str=new char[strlen("GETPLAYERNAME (  )")+strlen(s1)+1];
			sprintf(str,"GETPLAYERNAME ( %s )",s1);
			delete []s1;
		}
		break;
	case VALUE_CMD_GETPLAYERTYPE:
		{
			char *s1;

			s1=vpar[0]->generate_string();
			str=new char[strlen("GETPLAYERTYPE (  )")+strlen(s1)+1];
			sprintf(str,"GETPLAYERTYPE ( %s )",s1);
			delete []s1;
		}
		break;
	case VALUE_CMD_NEW_VIEWPORT:
		{
			char *s1,*s2,*s3,*s4,*s5,*s6;

			s1=vpar[0]->generate_string();
			s2=vpar[1]->generate_string();
			s3=vpar[2]->generate_string();
			s4=vpar[3]->generate_string();
			s5=vpar[4]->generate_string();
			s6=vpar[5]->generate_string();
			str=new char[strlen("NEW_VIEWPORT (  ,  ,  ,  ,  ,  )")+strlen(s1)+strlen(s2)+strlen(s3)+strlen(s4)+strlen(s5)+strlen(s6)+1];
			sprintf(str,"NEW_VIEWPORT ( %s , %s , %s , %s , %s , %s )",s1,s2,s3,s4,s5,s6);
			delete []s1;
			delete []s2;
			delete []s3;
			delete []s4;
			delete []s5;
			delete []s6;
		}
		break;
	case VALUE_CMD_ATOI:
		{
			char *s1;

			s1=vpar[0]->generate_string();
			str=new char[strlen("ATOI (  )")+strlen(s1)+1];
			sprintf(str,"ATOI ( %s )",s1);
			delete []s1;
		}
		break;
	case VALUE_CMD_ITOA:
		{
			char *s1,*s2;

			s1=vpar[0]->generate_string();
			s2=vpar[1]->generate_string();
			str=new char[strlen("ITOA (  ,  )")+strlen(s1)+strlen(s2)+1];
			sprintf(str,"ITOA ( %s , %s )",s1,s2);
			delete []s1;
			delete []s2;
		}
		break;
	case VALUE_CMD_DOT:
		{
			char *s1,*s2;

			s1=vpar[0]->generate_string();
			s2=vpar[1]->generate_string();
			str=new char[strlen("DOT (  ,  )")+strlen(s1)+strlen(s2)+1];
			sprintf(str,"DOT ( %s , %s )",s1,s2);
			delete []s1;
			delete []s2;
		}
		break;

	case VALUE_CMD_NEWPLAYER:
		{
			/* NEW_PLAYER ( obj , x , y , map ) */
			char *s,*s2,*s3,*s4;
			int len=0;

			s=vpar[0]->generate_string();
			len+=strlen(s);

			s2=vpar[1]->generate_string();
			len+=strlen(s2);

			s3=vpar[2]->generate_string();
			len+=strlen(s3);
			
			s4=vpar[3]->generate_string();
			len+=strlen(s4);

			str=new char[strlen("NEW_PLAYER (  ,  ,  ,  )")+len+1];
			sprintf(str,"NEW_PLAYER ( %s , %s , %s , %s )",s,s2,s3,s4);
			delete []s;
			delete []s2;
			delete []s3;
			delete []s4;
		}
		break;

	case VALUE_CMD_GET_FOCUSING:
		{
			/* int GET_FOCUSING */
			str=new char[strlen("GET_FOCUSING")+1];
			sprintf(str,"GET_FOCUSING");
		}
		break;

	case VALUE_CMD_GET_NFOCUSING:
		{
			/* int GET_NFOCUSING */
			str=new char[strlen("GET_NFOCUSING")+1];
			sprintf(str,"GET_NFOCUSING");
		}
		break;

	case VALUE_CMD_DISTANCE:
		{
			/* DISTANCE ( obj , "X"/"Y"/"XY" ) */
			char *s,*s2;
			int len=0;

			s=vpar[0]->generate_string();
			len+=strlen(s);

			s2=vpar[1]->generate_string();
			len+=strlen(s2);

			str=new char[strlen("DISTANCE (  ,  )")+len+1];
			sprintf(str,"DISTANCE ( %s , %s )",s,s2);
			delete []s;
			delete []s2;
		}
		break;

	} /* switch */ 

	return str;
} /* ValueSCRIPT::generate_string */ 


bool ValueSCRIPT::load(FILE *fp,MOG2 *game)
{
	char tmp[80];

	
	if (1!=fscanf(fp,"%s",tmp)) return false;

	if (strcmp(tmp,"INTEGER")==0) {
		command=VALUE_CMD_INTEGER;

		if (1!=fscanf(fp,"%i",&ipar)) return false;
	} else 	if (strcmp(tmp,"KEY")==0) {
		command=VALUE_CMD_KEY;

		if (1!=fscanf(fp,"%i",&ipar)) return false;
	} else 	if (strcmp(tmp,"STRING")==0) {
		command=VALUE_CMD_STRING;
		delete spar;
		if (!load_string(fp,&spar)) return false;
	} else 	if (strcmp(tmp,"VARIABLE")==0) {
		command=VALUE_CMD_VARIABLE;
		if (1!=fscanf(fp,"%i",&vlocation)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		delete vname;
		vname=new char[strlen(tmp)+1];
		strcpy(vname,tmp);
	} else 	if (strcmp(tmp,"RANDOM")==0) {
		command=VALUE_CMD_RANDOM;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[1]=new ValueSCRIPT();
		if (!vpar[1]->load(fp,game)) return false;	
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"ADD")==0) {
		command=VALUE_CMD_ADD;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[1]=new ValueSCRIPT();
		if (!vpar[1]->load(fp,game)) return false;	
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"SUB")==0) {
		command=VALUE_CMD_SUB;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[1]=new ValueSCRIPT();
		if (!vpar[1]->load(fp,game)) return false;	
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"MUL")==0) {
		command=VALUE_CMD_MUL;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[1]=new ValueSCRIPT();
		if (!vpar[1]->load(fp,game)) return false;	
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"DIV")==0) {
		command=VALUE_CMD_DIV;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[1]=new ValueSCRIPT();
		if (!vpar[1]->load(fp,game)) return false;	
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"ABS")==0) {
		command=VALUE_CMD_ABS;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"COUNT")==0) {
		command=VALUE_CMD_COUNT;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[1]=new ValueSCRIPT();
		if (!vpar[1]->load(fp,game)) return false;	
		if (1!=fscanf(fp,"%s",tmp)) return false;
		cond=new ConditionSCRIPT();
		if (!cond->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"COUNTCHILDREN")==0) {
		command=VALUE_CMD_COUNTCHILDREN;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		cond=new ConditionSCRIPT();
		if (!cond->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else if (strcmp(tmp,"CONCATENATE")==0) {
		command=VALUE_CMD_CONCATENATE;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[1]=new ValueSCRIPT();
		if (!vpar[1]->load(fp,game)) return false;	
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else if (strcmp(tmp,"REMOVE_SUBSTRING")==0) {
		command=VALUE_CMD_REMOVE_SUBSTRING;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[1]=new ValueSCRIPT();
		if (!vpar[1]->load(fp,game)) return false;	
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"MAKE_VARIABLE")==0) {
		command=VALUE_CMD_MAKE_VARIABLE;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[1]=new ValueSCRIPT();
		if (!vpar[1]->load(fp,game)) return false;	
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"KEYPRESS_ASCII")==0) {
		command=VALUE_CMD_KEYPRESS_ASCII;
	} else 	if (strcmp(tmp,"KEYPRESS_KEY")==0) {
		command=VALUE_CMD_KEYPRESS_KEY;
	} else 	if (strcmp(tmp,"KEYPRESS_NAME")==0) {
		command=VALUE_CMD_KEYPRESS_NAME;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"SAMPLE")==0) {
		command=VALUE_CMD_SAMPLE;
		delete []spar;
		if (!load_string(fp,&spar)) return false;
		sample=0;
		{
			SAMPLE *s2;
			List<SAMPLE> l;

			l.Instance(game->samples);
			l.Rewind();
			while(l.Iterate(s2)) {
				if (s2->cmp(spar)) sample=s2;
			} /* while */ 
		}
		delete []spar;
		spar=0;
	} else 	if (strcmp(tmp,"GETNPLAYERS")==0) {
		command=VALUE_CMD_GETNPLAYERS;
	} else 	if (strcmp(tmp,"GETPLAYERNAME")==0) {
		command=VALUE_CMD_GETPLAYERNAME;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"GETPLAYERTYPE")==0) {
		command=VALUE_CMD_GETPLAYERTYPE;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"NEW_VIEWPORT")==0) {
		command=VALUE_CMD_NEW_VIEWPORT;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[1]=new ValueSCRIPT();
		if (!vpar[1]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[2]=new ValueSCRIPT();
		if (!vpar[2]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[3]=new ValueSCRIPT();
		if (!vpar[3]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[4]=new ValueSCRIPT();
		if (!vpar[4]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[5]=new ValueSCRIPT();
		if (!vpar[5]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"ITOA")==0) {
		command=VALUE_CMD_ITOA;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[1]=new ValueSCRIPT();
		if (!vpar[1]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"ATOI")==0) {
		command=VALUE_CMD_ATOI;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"DOT")==0) {
		command=VALUE_CMD_DOT;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[1]=new ValueSCRIPT();
		if (!vpar[1]->load(fp,game)) return false;	
		if (1!=fscanf(fp,"%s",tmp)) return false;

	} else if (strcmp(tmp,"NEW_PLAYER")==0) {
		command=VALUE_CMD_NEWPLAYER;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		delete vpar[0];
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[1]=new ValueSCRIPT();
		if (!vpar[1]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[2]=new ValueSCRIPT();
		if (!vpar[2]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[3]=new ValueSCRIPT();
		if (!vpar[3]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
	} else 	if (strcmp(tmp,"GET_FOCUSING")==0) {
		command=VALUE_CMD_GET_FOCUSING;
	} else 	if (strcmp(tmp,"GET_NFOCUSING")==0) {
		command=VALUE_CMD_GET_NFOCUSING;

	} else 	if (strcmp(tmp,"DISTANCE")==0) {
		command=VALUE_CMD_DISTANCE;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[0]=new ValueSCRIPT();
		if (!vpar[0]->load(fp,game)) return false;
		if (1!=fscanf(fp,"%s",tmp)) return false;
		vpar[1]=new ValueSCRIPT();
		if (!vpar[1]->load(fp,game)) return false;	
		if (1!=fscanf(fp,"%s",tmp)) return false;

	} else {
		return false;
	} /* if */ 

	return true;
} /* ValueSCRIPT::load */ 


bool ValueSCRIPT::save(FILE *fp)
{
	char *str;

	str=generate_string();

	fprintf(fp,"%s\n",str);

	delete []str;
	return true;
} /* ValueSCRIPT::save */ 




void ValueSCRIPT::init(void)
{

} /* ValueSCRIPT::init */ 


int ValueSCRIPT::cycle(int *ival,char **sval,ActiveOBJECT **oval,ActiveOBJECT *o,ActiveROOM *r,ActiveMAP *m,ActiveMOG2 *g,ActiveINTERFACE *interf,ActiveMOG2Application *app,ActiveOBJECT **foreign_object,int location,KEYBOARDSTATE **LVK,int n_LVKs)
{

	switch(command) {
	case VALUE_CMD_INTEGER:
	case VALUE_CMD_KEY:
		if (ival!=0) *ival=ipar;
		return VALUE_INTEGER;
		break;
	case VALUE_CMD_STRING:
		if (sval!=0) *sval=spar;
		return VALUE_STRING;
		break;
	case VALUE_CMD_VARIABLE:

		switch(vlocation) {
		case VALUE_VAR_OBJECT:
			if (o==0) return VALUE_NONE;
			return o->get_variable_value(vname,ival,sval,oval);
			break;
		case VALUE_VAR_ROOM:
			if (r==0) return VALUE_NONE;
			return r->get_variable_value(vname,ival,sval,oval);
			break;
		case VALUE_VAR_MAP:
			if (m==0) return VALUE_NONE;
			return m->get_variable_value(vname,ival,sval,oval);
			break;
		case VALUE_VAR_GAME:
			if (g==0) return VALUE_NONE;
			return g->get_variable_value(vname,ival,sval,oval);
			break;
		case VALUE_VAR_INTERFACE:
			if (interf==0) return VALUE_NONE;
			return interf->get_variable_value(vname,ival,sval,oval);
			break;
		case VALUE_VAR_APPLICATION:
			if (app==0) return VALUE_NONE;
			return app->get_variable_value(vname,ival,sval,oval);
			break;
		case VALUE_VAR_FOREIGN:
			if (*foreign_object!=0) {

				return (*foreign_object)->get_variable_value(vname,ival,sval,oval);
			} /* if */ 

			return VALUE_NONE;
			break;
		} /* switch */ 
		return VALUE_NONE;
		break;

	case VALUE_CMD_RANDOM:
		{
			int v1,v2;
			if (vpar[0]->cycle(&v1,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER &&
				vpar[1]->cycle(&v2,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER) {
				if (v2>v1) *ival=app->random(v1,v2);
					  else *ival=app->random(v2,v1);
//				if (v2>v1) *ival=(rand()%((v2+1)-v1))+v1;
//					  else *ival=(rand()%((v1+1)-v2))+v2;

				return VALUE_INTEGER;
			} /* if */ 
		}
		break;
	case VALUE_CMD_ADD:
		{
			int v1,v2;
			if (vpar[0]->cycle(&v1,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER &&
				vpar[1]->cycle(&v2,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER) {
				*ival=v1+v2;
				return VALUE_INTEGER;
			} /* if */ 
		}
		break;
	case VALUE_CMD_SUB:
		{
			int v1,v2;
			if (vpar[0]->cycle(&v1,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER &&
				vpar[1]->cycle(&v2,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER) {
				*ival=v1-v2;
				return VALUE_INTEGER;
			} /* if */ 
		}
		break;
	case VALUE_CMD_MUL:
		{
			int v1,v2;
			if (vpar[0]->cycle(&v1,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER &&
				vpar[1]->cycle(&v2,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER) {
				*ival=v1*v2;
				return VALUE_INTEGER;
			} /* if */ 
		}
		break;
	case VALUE_CMD_DIV:
		{
			int v1,v2;
			if (vpar[0]->cycle(&v1,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER &&
				vpar[1]->cycle(&v2,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER) {
				*ival=v1/v2;
				return VALUE_INTEGER;
			} /* if */ 
		}
		break;
	case VALUE_CMD_ABS:
		{
			int v1;
			if (vpar[0]->cycle(&v1,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER) {
				*ival=abs(v1);
				return VALUE_INTEGER;
			} /* if */ 
		}
		break;
	case VALUE_CMD_COUNT:
		{
			char *s1,*s2;
			if (vpar[0]->cycle(0,&s1,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_STRING &&
				vpar[1]->cycle(0,&s2,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_STRING) {
				*ival=0;
				if (strcmp(s2,"room")==0) {
					int i;
					int count=0;
					List<ActiveOBJECT> l;
					ActiveOBJECT *obj;

					for(i=0;i<MAX_OBJECT_PLANES;i++) {
					    l.Instance(r->objects[i]);
					    l.Rewind();
					    while(l.Iterate(obj)) {
						count+=obj->count_objects(s1,cond,o,r,m,g,interf,app,location);
					    } /* while */ 
					} /* for */ 

					*ival=count;
				} /* if */ 
				if (strcmp(s2,"map")==0) {
					int i;
					int count=0;	
					List<ActiveROOM> l2;
					ActiveROOM *room;
					List<ActiveOBJECT> l;
					ActiveOBJECT *obj;

					l2.Instance(m->active_rooms);
					l2.Rewind();
					while(l2.Iterate(room)) {
					    for(i=0;i<MAX_OBJECT_PLANES;i++) {
						l.Instance(room->objects[i]);
						l.Rewind();
						while(l.Iterate(obj)) {
							count+=obj->count_objects(s1,cond,o,r,m,g,interf,app,location);
						} /* while */ 
					    } /* for */ 

					} /* while */ 

					*ival=count;
				} /* if */ 
				if (strcmp(s2,"game")==0) {
					int i;
					int count=0;	
					List<ActiveMAP> l3;
					ActiveMAP *map;
					List<ActiveROOM> l2;
					ActiveROOM *room;
					List<ActiveOBJECT> l;
					ActiveOBJECT *obj;

					l3.Instance(g->active_maps);
					l3.Rewind();
					while(l3.Iterate(map)) {
						l2.Instance(map->active_rooms);
						l2.Rewind();
						while(l2.Iterate(room)) {
						    for(i=0;i<MAX_OBJECT_PLANES;i++) {
							l.Instance(room->objects[i]);
							l.Rewind();
							while(l.Iterate(obj)) {
								count+=obj->count_objects(s1,cond,o,r,m,g,interf,app,location);
							} /* while */ 
						    } /* for */ 
						} /* while */ 
					} /* while */ 

					*ival=count;
				} /* if */ 
				return VALUE_INTEGER;
			} /* if */ 
		}
		break;
	case VALUE_CMD_COUNTCHILDREN:
		{
			char *s1;
			if (vpar[0]->cycle(0,&s1,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_STRING) {
				*ival=o->count_objects(s1,cond,o,r,m,g,interf,app,location);
				return VALUE_INTEGER;
			} /* if */ 
		}
		break;
	case VALUE_CMD_CONCATENATE:
		{
			char *s1,*s2;
			if (vpar[0]->cycle(0,&s1,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_STRING &&
				vpar[1]->cycle(0,&s2,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_STRING) {
				char *tmp;

				if (s2!=0 && (strcmp(s2,"\b")==0 || strcmp(s2,"\x7f")==0)) {
					int i;
					if (s1==0 || strlen(s1)==0) {
						tmp=new char[1];
						tmp[0]=0;
						app->tmp_strings.Add(tmp);
						*sval=tmp;
					} else {
						tmp=new char[strlen(s1)];
						for(i=0;i<int(strlen(s1));i++) tmp[i]=s1[i];
						tmp[strlen(s1)-1]=0;
						app->tmp_strings.Add(tmp);
						*sval=tmp;
					} /* if */ 
				} else {
					int l1=0,l2=0;

					if (s1!=0) l1=strlen(s1);
					if (s2!=0) l2=strlen(s2);
					tmp=new char[l1+l2+1];
					if (s1!=0 && s2!=0) {
						sprintf(tmp,"%s%s",s1,s2);
					} else {
						if (s1!=0) {
							sprintf(tmp,"%s",s1);
						} else {
							if (s2!=0) {
								sprintf(tmp,"%s",s2);
							} else {
								tmp[0]=0;
							} /* if */ 
						} /* if */ 
					} /* if */ 

					app->tmp_strings.Add(tmp);
					*sval=tmp;
				} /* if */ 

				return VALUE_STRING;
			} /* if */ 
		}
		break;
	case VALUE_CMD_REMOVE_SUBSTRING:
		{
			char *s1,*s2;
			if (vpar[0]->cycle(0,&s2,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_STRING &&
				vpar[1]->cycle(0,&s1,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_STRING) {
				char *tmp;
				int i=0,j=0,where=0;
				bool found=false;

				for(i=0;s1[i]!=0 && !found;i++) {
					for(j=0;s2[j]!=0 && s1[i+j]==s2[j];j++);
					if (s2[j]==0) {
						where=i;
						found=true;					
					} /* for */ 
				} /* for */ 

				if (found) {
					tmp=new char[strlen(s1)+1-strlen(s2)];
					for(i=0;i<where;i++) {
						tmp[i]=s1[i];
					} /* for */ 
					for(;s1[i+strlen(s2)]!=0;i++) {
						tmp[i]=s1[i+strlen(s2)];
					} /* for */ 
					tmp[i]=0;

					app->tmp_strings.Add(tmp);
					*sval=tmp;
				} else {
					*sval=s1;
				} /* if */ 

				return VALUE_STRING;
			} /* if */ 
		}
		break;
	case VALUE_CMD_MAKE_VARIABLE:
		char *s1,*s2;
		if (vpar[0]->cycle(0,&s1,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_STRING &&
			vpar[1]->cycle(0,&s2,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_STRING) {
			if (strcmp(s2,"object")==0) return o->get_variable_value(s1,ival,sval,oval);
			if (strcmp(s2,"room")==0) return r->get_variable_value(s1,ival,sval,oval);
			if (strcmp(s2,"map")==0) return m->get_variable_value(s1,ival,sval,oval);
			if (strcmp(s2,"game")==0) return g->get_variable_value(s1,ival,sval,oval);
			if (strcmp(s2,"interface")==0 && interf!=0) return interf->get_variable_value(s1,ival,sval,oval);
			if (strcmp(s2,"application")==0 && app!=0) return app->get_variable_value(s1,ival,sval,oval);
			if (strcmp(s2,"foreign")==0 && *foreign_object!=0) return (*foreign_object)->get_variable_value(s1,ival,sval,oval);
			if (s2[0]=='i' &&
				s2[1]=='n' &&
				s2[2]=='t' &&
				s2[3]=='e' &&
				s2[4]=='r' &&
				s2[5]=='f' &&
				s2[6]=='a' &&
				s2[7]=='c' &&
				s2[8]=='e' &&
				s2[9]==' ') {
				int i;

				if (1==sscanf(s2+10,"%i",&i)) {
					if (i>=0 && i<g->viewports.Length()) {
						return g->viewports[i]->current_interface->get_variable_value(s1,ival,sval,oval);
					} /* if */ 
				} /* if */ 
			} /* if */ 


			return VALUE_NONE;
		} /* if */ 
		break;
	case VALUE_CMD_KEYPRESS_ASCII:
		{
			List<SDL_keysym> l;
			SDL_keysym *ks;

			if (o==0 || o->LVK_ID==-1) {
				int j;
				for(j=0;j<n_LVKs;j++) {
					l.Instance(LVK[j]->keyevents);
					l.Rewind();
					while(l.Iterate(ks)) {
						if ((ks->unicode & 0xFF80) == 0) {
							*sval=new char[2];
							(*sval)[0]=ks->unicode & 0x7F;
							(*sval)[1]=0;
							return VALUE_STRING;
						} /* if */ 
					} /* while */ 
				} /* for */ 
			} else {
				l.Instance(LVK[o->LVK_ID]->keyevents);
				l.Rewind();
				while(l.Iterate(ks)) {
					if ((ks->unicode & 0xFF80) == 0) {
						*sval=new char[2];
						(*sval)[0]=ks->unicode & 0x7F;
						(*sval)[1]=0;
						return VALUE_STRING;
					} /* if */ 
				} /* while */ 
			} /* if */ 
			*sval=new char[1];
			**sval=0;
			return VALUE_STRING;
		}
		break;
	case VALUE_CMD_KEYPRESS_KEY:
		{
			List<SDL_keysym> l;
			SDL_keysym *ks;

			if (o==0 || o->LVK_ID==-1) {
				int j,i;
				for(j=0;j<n_LVKs;j++) {
					l.Instance(LVK[j]->keyevents);
					l.Rewind();
					while(l.Iterate(ks)) {
						if ((ks->unicode & 0xFF80) == 0) {
							(*ival)=ks->sym;
							return VALUE_INTEGER;
						} /* if */ 
					} /* while */ 
				} /* for */ 

				for(j=0;j<LVK[j]->n_joysticks*12;j++) {
					for(i=0;i<LVK[j]->n_joysticks*12;i++) {
						if (LVK[j]->keyboard[LVK[j]->joystick_0_pos+i] &&
							!LVK[j]->old_keyboard[LVK[j]->joystick_0_pos+i]) {
							*ival=LVK[j]->joystick_0_pos+i;
							return VALUE_INTEGER;
						} /* if */ 
					} /* for */ 
				} /* for */ 

			} else {
				l.Instance(LVK[o->LVK_ID]->keyevents);
				l.Rewind();
				while(l.Iterate(ks)) {
					if ((ks->unicode & 0xFF80) == 0) {
						(*ival)=ks->unicode & 0x7F;
						return VALUE_INTEGER;
					} /* if */ 
				} /* while */ 

				int i;
				for(i=0;i<LVK[o->LVK_ID]->n_joysticks*12;i++) {
					if (LVK[o->LVK_ID]->keyboard[LVK[o->LVK_ID]->joystick_0_pos+i] &&
						!LVK[o->LVK_ID]->old_keyboard[LVK[o->LVK_ID]->joystick_0_pos+i]) {
						*ival=LVK[o->LVK_ID]->joystick_0_pos+i;
						return VALUE_INTEGER;
					} /* if */ 
				} /* for */ 
			} /* if */ 
			return VALUE_NONE;
		}
		break;
	case VALUE_CMD_KEYPRESS_NAME:
		{
			int i1;
			if (vpar[0]->cycle(&i1,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER) {
				*sval=input_key_name(i1);
				return VALUE_STRING;
			} /* if */ 
			return VALUE_NONE;
		}
		break;
	case VALUE_CMD_SAMPLE:
		if (sample!=0) *sval=sample->name;
				  else *sval=0;
		return VALUE_STRING;
		break;

	case VALUE_CMD_GETNPLAYERS:
		if (g!=0) {
			*ival=g->player_objects.Length();
			return VALUE_INTEGER;
		} /* if */ 
		return VALUE_NONE;
		break;

	case VALUE_CMD_GETPLAYERNAME:
		{
			int i1;
			if (vpar[0]->cycle(&i1,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER) {
				if (g!=0) {
					if (i1>=0 && i1<g->player_names.Length()) {
						*sval=g->player_names[i1];
						return VALUE_STRING;
					} /* if */ 
				} /* if */ 
			} /* if */ 
				
			return VALUE_NONE;
		}
		break;

	case VALUE_CMD_GETPLAYERTYPE:
		{
			int i1;
			if (vpar[0]->cycle(&i1,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER) {
				if (g!=0) {
					if (i1>=0 && i1<g->player_objects.Length()) {
						*sval=g->player_objects[i1]->object->name;
						return VALUE_STRING;
					} /* if */ 
				} /* if */ 
			} /* if */ 
				
			return VALUE_NONE;
		}
		break;

	case VALUE_CMD_NEW_VIEWPORT:
		{
			int x,y,w,h,zoom;
			char *interf_name;

			if (vpar[0]->cycle(&x,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER &&
				vpar[1]->cycle(&y,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER &&
				vpar[2]->cycle(&w,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER &&
				vpar[3]->cycle(&h,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER &&
				vpar[4]->cycle(&zoom,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER &&
				vpar[5]->cycle(0,&interf_name,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_STRING) {
				VIEWPORT *v;
				SDL_Rect vp;
				List<MOG2_INTERFACE> l;
				MOG2_INTERFACE *i,*interf=0;
				bool found=false;

				l.Instance(g->active_game->interfaces);
				l.Rewind();
				while(!found && l.Iterate(i)) {
					if (i->cmp(interf_name)) {
						interf=i;
						found=true;
					} /* if */ 
				} /* while */ 

				vp.x=x;
				vp.y=y;
				vp.w=w;
				vp.h=h;

				v=new VIEWPORT();
				v->vp=vp;
				v->current_interface=0;
				v->zoom=float(zoom)/100.0F;

				if (interf!=0) {
					v->current_interface=new ActiveINTERFACE(interf,g);
				} /* if */ 

				g->viewports.Add(v);
				*ival=g->viewports.Length()-1;
				return VALUE_INTEGER;
			} /* if */ 

		}
		return VALUE_NONE;
		break;

	case VALUE_CMD_ITOA:
		{
			int i,ndigits;
			if (vpar[0]->cycle(&i,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER &&
				vpar[1]->cycle(&ndigits,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER) {
				char pretmp[100],tmp[100],*tmp2;

				sprintf(pretmp,"%c%c%i%c",'%','.',ndigits,'i');
				sprintf(tmp,pretmp,i);
				tmp2=new char[strlen(tmp)+1];
				strcpy(tmp2,tmp);
				*sval=tmp2;
				app->tmp_strings.Add(tmp2);
				return VALUE_STRING;
			} /* if */ 
		}		
		return VALUE_NONE;
		break;

	case VALUE_CMD_ATOI:
		{
			char *s;

			if (vpar[0]->cycle(0,&s,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_STRING) {
				int i;

				if (1==sscanf(s,"%i",&i)) {
					*ival=i;
					return VALUE_INTEGER;
				} /* if */ 
			} /* if */ 	
		} 
		return VALUE_NONE;
		break;

	case VALUE_CMD_DOT:
		{
			ActiveOBJECT *otmp=0;
			char *stmp=0;

			if (vpar[0]->cycle(0,0,&otmp,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_OBJECT &&
				vpar[1]->cycle(0,&stmp,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_STRING) {

				if (otmp!=0) {
					if (app!=0) {
						if (!app->find_object(otmp)) otmp=0;
					} else {
						if (g!=0) {
							if (!g->find_object(otmp)) otmp=0;
						} /* if */ 
					} /* if */ 
					if (otmp!=0) {
						return (otmp)->get_variable_value(stmp,ival,sval,oval);
					} /* if */ 
				} /* if */ 

				return VALUE_NONE;
			} /* if */ 
		}
		break;

	case VALUE_CMD_NEWPLAYER:
		{
			char *object_name;
			char *map_name;
			int x,y;
			if (vpar[0]->cycle(0,&object_name,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_STRING &&
				vpar[1]->cycle(&x,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER &&
				vpar[2]->cycle(&y,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER &&
				vpar[3]->cycle(0,&map_name,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_STRING) {

				OBJECT *target_o=g->find_object(object_name);
				ActiveOBJECT *ao;
				MAP *target_map=0;
				ROOM *target_room=0;

				if (target_o!=0) {
					/* look for the destination map: */ 
					List<ActiveMAP> l;
					ActiveMAP *am;
					List<MAP> l2;
					MAP *m;
					bool found=false;

					l.Instance(g->active_maps);
					l.Rewind();
					while(!found && l.Iterate(am)) {
						if (am->map->cmp(map_name)) {
							target_map=am->map;
							found=true;
						} /* if */ 
					} /* while */ 

					if (!found) {
						l2.Instance(g->active_game->maps);
						l2.Rewind();
						while(!found && l2.Iterate(m)) {
							if (m->cmp(map_name)) {
								target_map=m;
								found=true;
							} /* if */ 
						} /* while */ 
					} /* if */ 

					if (target_map!=0) {
						/* Look for the destination room: */ 
						target_room=target_map->get_room(x/TILESIZE,y/TILESIZE);
						
						if (target_room!=0) {
							ActiveROOM *r;

							r=g->add_room(target_room);

							ao=new ActiveOBJECT(0,0,-1,target_o,r,0);
							if (app!=0) {
								r->objects[app->player_plane].Add(ao);
								r->nonpassive_objects[app->player_plane].Add(ao);
							} else {
								int i=0;
								List<VARIABLE> l;
								VARIABLE *var;

								l.Instance(g->active_game->application_variables);
								l.Rewind();
								while(l.Iterate(var)) {
									if (var->cmp("player_plane")) {
										var->get(&i,0,0);
									} /* if */ 
								} /* while */ 
								r->objects[i].Add(ao);
								r->nonpassive_objects[i].Add(ao);
							} /* if */ 

							ao->x=ao->old_x=x-(target_room->x*TILESIZE);
							ao->y=ao->old_y=y-(target_room->y*TILESIZE);
							ao->LVK_ID=0;

							g->player_objects.Add(ao);
							g->player_names.Add(0);
						} /* if */ 
					} /* if */ 

				} /* if */ 

				*foreign_object=ao;
				*ival=g->player_objects.Length()-1;
				return VALUE_INTEGER;
			} /* if */ 
		}
		break;

	case VALUE_CMD_GET_FOCUSING:
		{
			int n;

			if (o->object->control==CTRL_GAMEWINDOW &&
				vpar[0]->cycle(&n,0,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_INTEGER) {
				if (n>=0 && n<o->gw_focusing.Length()) {
					*oval=o->gw_focusing[n];
					return VALUE_OBJECT;
				} else {
					return VALUE_NONE;
				} /* if */ 
			} /* if */ 

			return VALUE_NONE;		
		}
		break;

	case VALUE_CMD_GET_NFOCUSING:
		if (o->object->control==CTRL_GAMEWINDOW) {
			*ival=o->gw_focusing.Length();
			return VALUE_INTEGER;
		} else {
			return VALUE_NONE;
		} /* if */ 
		break;

	case VALUE_CMD_DISTANCE:
		{
			ActiveOBJECT *o2;
			char *type;

			if (vpar[0]->cycle(0,0,&o2,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_OBJECT &&
				vpar[1]->cycle(0,&type,0,o,r,m,g,interf,app,foreign_object,location,LVK,n_LVKs)==VALUE_STRING) {

				if (o!=0 && o2!=0) {
					int ax1,ay1;
					int ax2,ay2;
					int t1,t2;
					int dif;

					o->absolute_hotspot_coordinates(&ax1,&ay1);
					o2->absolute_hotspot_coordinates(&ax2,&ay2);

					t1=ax1;
					t2=ax2;
					dif=abs(t1-t2);
					if (o2->room->room->horizontal_circularity) {
						if (abs(ax1-(ax2+o2->room->room->get_dx()*TILESIZE))<dif) {
							t1=ax1;
							t2=ax2+o2->room->room->get_dx()*TILESIZE;
							dif=abs(t1-t2);
						} /* if */ 
						if (abs(ax1-(ax2-o2->room->room->get_dx()*TILESIZE))<dif) {
							t1=ax1;
							t2=ax2-o2->room->room->get_dx()*TILESIZE;
							dif=abs(t1-t2);
						} /* if */ 
					} /* if */ 

					if (o->room->room->horizontal_circularity) {
						if (abs((ax1+o->room->room->get_dx()*TILESIZE)-ax2)<dif) {
							t1=(ax1+o->room->room->get_dx()*TILESIZE);
							t2=ax2;
							dif=abs(t1-t2);
						} /* if */ 
						if (abs((ax1-o->room->room->get_dx()*TILESIZE)-ax2)<dif) {
							t1=(ax1-o->room->room->get_dx()*TILESIZE);
							t2=ax2;
							dif=abs(t1-t2);
						} /* if */ 

						if (o2->room->room->horizontal_circularity) {
							if (abs((ax1+o->room->room->get_dx()*TILESIZE)-(ax2+o2->room->room->get_dx()*TILESIZE))<dif) {
								t1=(ax1+o->room->room->get_dx()*TILESIZE);
								t2=ax2+o2->room->room->get_dx()*TILESIZE;
								dif=abs(t1-t2);
							} /* if */ 
							if (abs((ax1+o->room->room->get_dx()*TILESIZE)-(ax2-o2->room->room->get_dx()*TILESIZE))<dif) {
								t1=(ax1+o->room->room->get_dx()*TILESIZE);
								t2=ax2-o2->room->room->get_dx()*TILESIZE;
								dif=abs(t1-t2);
							} /* if */ 
							
							if (abs((ax1-o->room->room->get_dx()*TILESIZE)-(ax2+o2->room->room->get_dx()*TILESIZE))<dif) {
								t1=(ax1-o->room->room->get_dx()*TILESIZE);
								t2=ax2+o2->room->room->get_dx()*TILESIZE;
								dif=abs(t1-t2);
							} /* if */ 
							if (abs((ax1-o->room->room->get_dx()*TILESIZE)-(ax2-o2->room->room->get_dx()*TILESIZE))<dif) {
								t1=(ax1-o->room->room->get_dx()*TILESIZE);
								t2=ax2-o2->room->room->get_dx()*TILESIZE;
								dif=abs(t1-t2);
							} /* if */ 
						} /* if */ 
					} /* if */ 
					ax1=t1;
					ax2=t2;

					t1=ay1;
					t2=ay2;
					dif=abs(t1-t2);
					if (o2->room->room->vertical_circularity) {
						if (abs(ay1-(ay2+o2->room->room->get_dy()*TILESIZE))<dif) {
							t1=ay1;
							t2=ay2+o2->room->room->get_dy()*TILESIZE;
							dif=abs(t1-t2);
						} /* if */ 
						if (abs(ay1-(ay2-o2->room->room->get_dy()*TILESIZE))<dif) {
							t1=ay1;
							t2=ay2-o2->room->room->get_dy()*TILESIZE;
							dif=abs(t1-t2);
						} /* if */ 
					} /* if */ 

					if (o->room->room->vertical_circularity) {
						if (abs((ay1+o->room->room->get_dy()*TILESIZE)-ay2)<dif) {
							t1=(ay1+o->room->room->get_dy()*TILESIZE);
							t2=ay2;
							dif=abs(t1-t2);
						} /* if */ 
						if (abs((ay1-o->room->room->get_dy()*TILESIZE)-ay2)<dif) {
							t1=(ay1-o->room->room->get_dy()*TILESIZE);
							t2=ay2;
							dif=abs(t1-t2);
						} /* if */ 

						if (o2->room->room->vertical_circularity) {
							if (abs((ay1+o->room->room->get_dy()*TILESIZE)-(ay2+o2->room->room->get_dy()*TILESIZE))<dif) {
								t1=(ay1+o->room->room->get_dy()*TILESIZE);
								t2=ay2+o2->room->room->get_dy()*TILESIZE;
								dif=abs(t1-t2);
							} /* if */ 
							if (abs((ay1+o->room->room->get_dy()*TILESIZE)-(ay2-o2->room->room->get_dy()*TILESIZE))<dif) {
								t1=(ay1+o->room->room->get_dy()*TILESIZE);
								t2=ay2-o2->room->room->get_dy()*TILESIZE;
								dif=abs(t1-t2);
							} /* if */ 
							
							if (abs((ay1-o->room->room->get_dy()*TILESIZE)-(ay2+o2->room->room->get_dy()*TILESIZE))<dif) {
								t1=(ay1-o->room->room->get_dy()*TILESIZE);
								t2=ay2+o2->room->room->get_dy()*TILESIZE;
								dif=abs(t1-t2);
							} /* if */ 
							if (abs((ay1-o->room->room->get_dy()*TILESIZE)-(ay2-o2->room->room->get_dy()*TILESIZE))<dif) {
								t1=(ay1-o->room->room->get_dy()*TILESIZE);
								t2=ay2-o2->room->room->get_dy()*TILESIZE;
								dif=abs(t1-t2);
							} /* if */ 
						} /* if */ 
					} /* if */ 
					ay1=t1;
					ay2=t2;

					if (strcmp(type,"x")==0 ||
						strcmp(type,"X")==0) {
						*ival=ax2-ax1;
						return VALUE_INTEGER;
					} else {
						if (strcmp(type,"y")==0 ||
							strcmp(type,"Y")==0) {
							*ival=ay2-ay1;
							return VALUE_INTEGER;
						} else {
							if (strcmp(type,"xy")==0 ||
								strcmp(type,"XY")==0 ||
								strcmp(type,"yx")==0 ||
								strcmp(type,"YX")==0) {
								*ival=int(sqrt((ax2-ax1)*(ax2-ax1)+(ay2-ay1)*(ay2-ay1)));
								return VALUE_INTEGER;
							} else {
							} /* if */ 
						} /* if */ 
					} /* if */ 
				} else {
					return VALUE_NONE;
				} /* if */ 
				
			} /* if */ 

			return VALUE_NONE;		
		}
		break;

	} /* switch */ 

	return VALUE_NONE;
} /* ValueSCRIPT::cycle */ 


void ValueSCRIPT::copy_state(ValueSCRIPT *s)
{
	int i;
	command=s->command;
	ipar=s->ipar;

	delete []spar;
	if (s->spar!=0) {
		spar=new char[strlen(s->spar)+1];
		strcpy(spar,s->spar);
	} else {
		spar=0;
	} /* if */ 

	vlocation=s->vlocation;

	delete []vname;
	if (s->vname!=0) {
		vname=new char[strlen(s->vname)+1];
		strcpy(vname,s->vname);
	} else {
		vname=0;
	} /* if */ 

	for(i=0;i<6;i++) {
		if (s->vpar[i]!=0) {
			if (vpar[i]==0) vpar[i]=new ValueSCRIPT();
			vpar[i]->copy_state(s->vpar[i]);
		} else {
			if (vpar[i]!=0) {
				delete vpar[i];
				vpar[i]=0;
			} /* if */ 
		} /* if */ 
	} /* for */ 
	sample=s->sample;

	if (s->cond!=0) {
		if (cond==0) cond=new ConditionSCRIPT();
		cond->copy_state(s->cond);
	} else {
		if (cond!=0) {
			delete cond;
			cond=0;
		} /* if */ 
	} /* if */ 

	dx=0;
	dy=0;
} /* ValueSCRIPT::copy_state */ 


void ValueSCRIPT::simplify(void)
{

} /* ValueSCRIPT::simplify */ 


bool ValueSCRIPT::uses_sample(SAMPLE *s)
{
	if (vpar[0]!=0) if (vpar[0]->uses_sample(s)) return true;
	if (vpar[1]!=0) if (vpar[1]->uses_sample(s)) return true;

	if (command==VALUE_CMD_SAMPLE && sample==s) return true;

	return false;
} /* ValueSCRIPT::uses_sample */ 
